"""
See Problem freepizza in MiniZinc

Example of Execution:
  python3 PizzaVoucher.py -data=PizzaVoucher_pizza6.json
"""

from pycsp3 import *

prices, vouchers = data
nPizzas, nVouchers = len(prices), len(vouchers)

# v[i] is the voucher used for the ith pizza. 0 means that no voucher is used.
# A negative (resp., positive) value i means that the ith pizza contributes to the the pay (resp., free) part of voucher |i|.
v = VarArray(size=nPizzas, dom=range(-nVouchers, nVouchers + 1))

# p[i] is the number of paid pizzas wrt the ith voucher
p = VarArray(size=nVouchers, dom=lambda i: {0, vouchers[i].payPart})

# f[i] is the number of free pizzas wrt the ith voucher
f = VarArray(size=nVouchers, dom=lambda i: range(vouchers[i].freePart + 1))

satisfy(
    # counting paid pizzas
    [Count(v, value=-i - 1) == p[i] for i in range(nVouchers)],

    # counting free pizzas
    [Count(v, value=i + 1) == f[i] for i in range(nVouchers)],

    # a voucher, if used, must contribute to have at least one free pizza.
    [iff(f[i] == 0, p[i] != vouchers[i].payPart) for i in range(nVouchers)],

    # a free pizza obtained with a voucher must be cheaper than any pizza paid wrt this voucher
    [imply(v[i] < 0, v[i] != -v[j]) for i in range(nPizzas) for j in range(nPizzas) if i != j and prices[i] < prices[j]]
)

minimize(
    # minimizing summed up costs of pizzas
    Sum((v[i] <= 0) * prices[i] for i in range(nPizzas))
)

annotate(decision=v)


#  [(f[i] == 0) == (p[i] != vouchers[i].payPart) for i in range(nVouchers)],  is this clearer?



"""
Problem 049 on CSPLib

Examples of Execution:
  python3 NumberPartitioning.py
  python3 NumberPartitioning.py -data=10
"""

from pycsp3 import *

n = data or 8
assert n % 2 == 0, "The value of n must be even"

# x[i] is the ith value of the first set
x = VarArray(size=n // 2, dom=range(1, n + 1))

# y[i] is the ith value of the second set
y = VarArray(size=n // 2, dom=range(1, n + 1))

satisfy(
    AllDifferent(x + y),

    # tag(power1)
    [
        Sum(x) == n * (n + 1) // 4,
        Sum(y) == n * (n + 1) // 4
    ],

    # tag(power2)
    [
        Sum(x[i] * x[i] for i in range(n // 2)) == n * (n + 1) * (2 * n + 1) // 12,
        Sum(y[i] * y[i] for i in range(n // 2)) == n * (n + 1) * (2 * n + 1) // 12
    ],

    # tag(symmetry-breaking)
    [
        x[0] == 1,
        Increasing(x, strict=True),
        Increasing(y, strict=True)
    ]
)



"""
Problem 041 on CSPLib

Execution:
  python3 NFractions.py
"""

from pycsp3 import *

digits = VarArray(size=9, dom=range(1, 10))
a, b, c, d, e, f, g, h, i = digits

satisfy(
    AllDifferent(digits),

    a * (10 * e + f) * (10 * h + i) + d * (10 * b + c) * (10 * h + i) + g * (10 * b + c) * (10 * e * f) == (10 * b + c) * (10 * e + f) * (10 * h + i)
)



from pycsp3.problems.data.parsing import *

data["nSteps"] = number_in(line())
data["nUsers"] = number_in(next_line())
nConstraints = number_in(next_line())


def auth(line):
    s = line[line.index(":") + 1:].strip()
    return [i for i in range(len(s)) if s[i] == '1']


next_line()
data["auths"] = [auth(next_line()) for _ in range(data["nUsers"])]
next_line()
am, al = [], []
for _ in range(nConstraints):
    line = next_line().lower().strip()
    t = numbers_in(line)
    if "least" in line:
        al.append((t[0], t[1:]))
    elif "most" in line:
        am.append((t[0], t[1:]))
    else:
        raise AssertionError
data["atMost"] = am
data["atLeast"] = al



#
# if next_line().startswith("capacity"):
#     data.warehouseCapacities = numbers_in(line())
#     next_line()
# else:
#     data.warehouseCapacities = [data.nStores] * data.nWarehouses
#
# data.storeSupplyCosts = [numbers for numbers in [numbers_in(line) for line in remaining_lines()] if numbers]



"""
Problem 004 on CSPLib

Example of Execution:
  python3 MisteryShopper.py -data=MisteryShopper_04.json
"""

from pycsp3 import *

vr_sizes = data.visitorGroups  # vr_sizes[i] gives the size of the ith visitor group
ve_sizes = data.visiteeGroups  # ve_sizes[i] gives the size of the ith visitee group

nVisitors, nVisitees = sum(vr_sizes), sum(ve_sizes)
assert nVisitors >= nVisitees, "The number of visitors must be greater than the number of visitees"
if nVisitors - nVisitees > 0:
    ve_sizes.append(nVisitors - nVisitees)  # an artificial group with dummy visitees is added
n, nWeeks = nVisitors, len(vr_sizes)

# r[i][w] is the visitor for the ith visitee at week w
r = VarArray(size=[n, nWeeks], dom=range(n))

# e[i][w] is the visitee for the ith visitor at week w
e = VarArray(size=[n, nWeeks], dom=range(n))

# rg[i][w] is the visitor group for the ith visitee at week w
rg = VarArray(size=[n, nWeeks], dom=range(len(vr_sizes)))

# eg[i][w] is the visitee group for the ith visitor at week w
eg = VarArray(size=[n, nWeeks], dom=range(len(ve_sizes)))

vr_table = {(i, sum(vr_sizes[:i]) + j) for i, size in enumerate(vr_sizes) for j in range(size)}
ve_table = {(i, sum(ve_sizes[:i]) + j) for i, size in enumerate(ve_sizes) for j in range(size)}

satisfy(
    # each week, all visitors must be different
    [AllDifferent(col) for col in columns(r)],

    # each week, all visitees must be different
    [AllDifferent(col) for col in columns(e)],

    # the visitor groups must be different for each visitee
    [AllDifferent(row) for row in rg],

    # the visitee groups must be different for each visitor
    [AllDifferent(row) for row in eg],

    # channeling arrays vr and ve, each week
    [Channel(r[:, w], e[:, w]) for w in range(nWeeks)],

    # tag(symmetry-breaking)
    [
        LexIncreasing(r, matrix=True),

        [Increasing([r[i][w] for i in range(nVisitees, n)], strict=True) for w in range(nWeeks)]
    ],

    # linking a visitor with its group
    [(rg[i][w], r[i][w]) in vr_table for i in range(n) for w in range(nWeeks)],

    # linking a visitee with its group
    [(eg[i][w], e[i][w]) in ve_table for i in range(n) for w in range(nWeeks)]
)



"""
Problem 005 on CSPLib

Examples of Execution:
  python3 LowAutocorrelation.py
  python3 LowAutocorrelation.py -data=16
"""

from pycsp3 import *

n = data or 8

# x[i] is the ith value of the sequence to be built.
x = VarArray(size=n, dom={-1, 1})

# y[k][i] is the ith product value required to compute the kth auto-correlation
y = VarArray(size=[n - 1, n - 1], dom=lambda k, i: {-1, 1} if i < n - k - 1 else None)

# c[k] is the value of the kth auto-correlation
c = VarArray(size=n - 1, dom=lambda k: range(-n + k + 1, n - k))

satisfy(
    [y[k][i] == x[i] * x[i + k + 1] for k in range(n - 1) for i in range(n - k - 1)],

    [Sum(y[k]) == c[k] for k in range(n - 1)]
)

minimize(
    # minimizing the sum of the squares of the auto-correlation
    Sum(c[k] * c[k] for k in range(n - 1))
)



import os
from collections import OrderedDict

from pycsp3.classes.auxiliary.conditions import Condition
from pycsp3.classes.auxiliary.ptypes import TypeVar, TypeCtr, TypeCtrArg, TypeXML, TypeConditionOperator, TypeRank
from pycsp3.classes.auxiliary.values import IntegerEntity
from pycsp3.classes.entities import EVarArray, ECtr, TypeNode, Node
from pycsp3.classes.main.domains import Domain
from pycsp3.classes.main.variables import Variable, VariableInteger
from pycsp3.tools.compactor import compact
from pycsp3.tools.utilities import is_1d_list, matrix_to_string, transitions_to_string, integers_to_string, table_to_string, flatten, is_matrix, error


class Diffs:
    """"
      Objects of this class are used to record the differences between two (or more) close constraints.
      This allows us to build groups in an automatic way.
    """

    fusion = None  # the current Diffs object memorizing the arguments that differ between two (or more) constraints

    @staticmethod
    def reset():
        Diffs.fusion = None

    def __init__(self, records=[]):
        self.argument_names = [name for (name, flag) in records]
        self.argument_flags = [flag for (name, flag) in records]

    def merge(self):
        if Diffs.fusion is None:
            Diffs.fusion = self  # this is the first time such an object is built (since the last reset); we simply record it
        else:  # the recorded object (fusion) and the last built one are compatible (we know that at time of calling); so, we simply update flags
            for i, name in enumerate(self.argument_names):
                index = Diffs.fusion.argument_names.index(name)  # note that self.argument_names must be be included in Diffs.fusion.argument_names
                Diffs.fusion.argument_flags[index] |= self.argument_flags[i]
                # for i in range(len(Diffs.fusion.argument_flags)): Diffs.fusion.argument_flags[i] |= self.argument_flags[i]


class ConstraintArgument:
    def __init__(self, name, content, attributes=[], content_compressible=True, content_ordered=False, lifted=False):
        assert isinstance(name, (TypeCtrArg, TypeXML)), str(name) + " " + str(type(name))
        self.name = name  # name of the argument
        self.attributes = attributes  # list of pairs (key, value) representing the attributes
        self.content = content  # content of the argument
        self.content_compressible = content_compressible  # indicates if we can try to make the content more compact
        self.content_ordered = content_ordered  # indicates if the content must be kept as it is (order is important)
        self.lifted = lifted  # indicates if the constraint is lifted to several lists (or sets); see specifications

    def __str__(self):
        return str(self.name) + str(self.content)


class Constraint:
    def __init__(self, name):
        self.name = name
        self.attributes = []
        self.arguments = OrderedDict()  # arguments of the constraint (such as list, supports, condition, ...)
        self.n_parameters = 0  # used when building abstract forms of constraints (e.g., in groups for %0 %1 %2 ...)

    def arg(self, name, content, *, attributes=[], content_compressible=True, content_ordered=False, lifted=False):
        self.arguments[name] = ConstraintArgument(name, content, attributes, content_compressible, content_ordered, lifted)
        return self

    def set_condition(self, operator, right_operand):
        return self.arg(TypeCtrArg.CONDITION, Condition.build_condition((operator, right_operand)))

    def set_value(self, new_value):
        return self.arg(TypeCtrArg.VALUE, new_value)

    def similar_structure(self, other):
        if type(self) != type(other) or self.attributes != other.attributes:
            return False
        args1, args2 = list(self.arguments.values()), list(other.arguments.values())
        if len(args1) != len(args2):
            return False
        if any(args1[i].attributes != args2[i].attributes or args1[i].lifted != args2[i].lifted for i in range(len(args1))):
            return False
        return any(type(args1[i].content) == type(args2[i].content) for i in range(len(args1)))

    def close_to(self, other):
        if not self.similar_structure(other):
            return False
        args1, args2 = list(self.arguments.values()), list(other.arguments.values())
        records = []
        for i in range(len(args1)):
            if str(args1[i].content) != str(args2[i].content):
                if args1[i].name == TypeCtrArg.CONDITION and args1[i].content.operator != args2[i].content.operator:
                    return False  # the operators are different in the two conditions
                b = hasattr(args1[i].content, '__len__') and hasattr(args2[i].content, '__len__') and len(args1[i].content) != len(args2[i].content)
                records.append((args1[i].name, b))
        if len(records) > 2:
            return False  # too many differences
        diffs = Diffs(records)
        if len(records) == 0:
            return diffs
        if Diffs.fusion and not all(
                        name in Diffs.fusion.argument_names for name in diffs.argument_names):  # diffs.argument_names != Diffs.fusion.argument_names:
            return False  # because arguments are not the same (so it is not possible to make a unique abstraction)
        if len([flag for flag in diffs.argument_flags if flag]) > 1:
            return False  # two arguments of different size cannot be currently abstracted
        return diffs

    def parameter_form(self, p):
        length = len(p) if isinstance(p, list) else 1
        self.n_parameters += length
        return " ".join("%" + str(v + self.n_parameters - length) for v in range(length))

    def __str__(self):
        return str(self.name) + ": " + " ".join(str(a) for a in self.attributes) + " ".join(str(v) for k, v in self.arguments.items())


class ConstraintUnmergeable(Constraint):
    def __init__(self, name):
        super().__init__(name)

    def close_to(self, other):
        return False


class ConstraintWithCondition(Constraint):
    def __init__(self, name):
        super().__init__(name)

    def min_possible_value(self):
        pass

    def max_possible_value(self):
        pass


''' Generic Constraints '''


class ConstraintIntension(Constraint):
    def __init__(self, tree):
        super().__init__(TypeCtr.INTENSION)
        self.arg(TypeCtrArg.FUNCTION, tree)

    def abstract_tree(self):
        return self.arguments[TypeCtrArg.FUNCTION].content.abstract_tree()

    def abstract_values(self):
        return self.arguments[TypeCtrArg.FUNCTION].content.abstract_values()

    def close_to(self, other):
        if not self.similar_structure(other):
            return False
        return Diffs() if self.abstract_tree() == other.abstract_tree() else False


class ConstraintExtension(Constraint):
    cache = dict()

    @staticmethod
    def caching(table):
        if len(table) == 0:
            return None
        arity = 1 if is_1d_list(table, (int, str)) else len(table[0])
        h = hash(tuple(table))
        if h not in ConstraintExtension.cache:
            if arity > 1:
                table.sort()

                tbl = None  # we remove redundant tuples, if present
                for i in range(len(table) - 1):
                    if table[i] != table[i + 1]:
                        if tbl:
                            tbl.append(table[i])
                    else:
                        if not tbl:
                            tbl = table[:i]
                if tbl:
                    tbl.append(table[-1])
                    table = tbl

                ConstraintExtension.cache[h] = table_to_string(table, parallel=os.name != 'nt')
            elif isinstance(table[0], int):
                ConstraintExtension.cache[h] = integers_to_string(table)
            else:
                ConstraintExtension.cache[h] = " ".join(v for v in sorted(table))
        return ConstraintExtension.cache[h]

    def __init__(self, scope, table, positive=True):
        super().__init__(TypeCtr.EXTENSION)
        assert is_1d_list(scope, Variable)
        assert len(table) == 0 or (len(scope) == 1 and (is_1d_list(table, int) or is_1d_list(table, str))) or (len(scope) > 1 and len(scope) == len(table[0]))
        self.arg(TypeCtrArg.LIST, scope, content_ordered=True)
        self.arg(TypeCtrArg.SUPPORTS if positive else TypeCtrArg.CONFLICTS, ConstraintExtension.caching(table), content_compressible=False)

    def close_to(self, other):
        if not self.similar_structure(other):
            return False
        if len(self.arguments[TypeCtrArg.LIST].content) != len(other.arguments[TypeCtrArg.LIST].content):
            return False
        if TypeCtrArg.SUPPORTS in self.arguments and self.arguments[TypeCtrArg.SUPPORTS].content != other.arguments[TypeCtrArg.SUPPORTS].content:
            return False
        elif TypeCtrArg.CONFLICTS in self.arguments and self.arguments[TypeCtrArg.CONFLICTS].content != other.arguments[TypeCtrArg.CONFLICTS].content:
            return False
        return Diffs([(TypeCtrArg.LIST, False)])


''' Constraints defined from languages'''


class ConstraintRegular(Constraint):
    def __init__(self, lst, automaton):
        super().__init__(TypeCtr.REGULAR)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True)
        self.arg(TypeCtrArg.TRANSITIONS, transitions_to_string(automaton.transitions))
        self.arg(TypeCtrArg.START, automaton.start)
        self.arg(TypeCtrArg.FINAL, automaton.final)

    def close_to(self, other):
        diffs = super().close_to(other)
        return False if not diffs or TypeCtrArg.TRANSITIONS in diffs.argument_names else diffs


class ConstraintMdd(Constraint):
    def __init__(self, lst, transitions):
        super().__init__(TypeCtr.MDD)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True)
        # TODO reordering transitions in order to guarantee to have:
        # - the root as the src of the first transition
        # - the terminal as the dst of the last transition
        # - no transition with a src occurring before it was reached
        self.arg(TypeCtrArg.TRANSITIONS, transitions_to_string(transitions))


''' Comparison-based Constraints '''


class ConstraintAllDifferent(Constraint):
    def __init__(self, lst, excepting):
        super().__init__(TypeCtr.ALL_DIFFERENT)
        self.arg(TypeCtrArg.LIST, lst)
        self.arg(TypeCtrArg.EXCEPT, excepting)


class ConstraintAllDifferentList(ConstraintUnmergeable):
    def __init__(self, lst, excepting):
        super().__init__(TypeCtr.ALL_DIFFERENT)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True, lifted=True)
        self.arg(TypeCtrArg.EXCEPT, excepting)


class ConstraintAllDifferentMatrix(ConstraintUnmergeable):
    def __init__(self, lst):
        super().__init__(TypeCtr.ALL_DIFFERENT)
        self.arg(TypeCtrArg.MATRIX, matrix_to_string(lst), content_compressible=lst)


class ConstraintAllEqual(Constraint):
    def __init__(self, lst):
        super().__init__(TypeCtr.ALL_EQUAL)
        self.arg(TypeCtrArg.LIST, lst)


class ConstraintOrdered(Constraint):
    def __init__(self, lst, operator, lengths):
        super().__init__(TypeCtr.ORDERED)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True)
        self.arg(TypeCtrArg.LENGTHS, lengths, content_ordered=True)
        self.arg(TypeCtrArg.OPERATOR, operator)


class ConstraintLex(ConstraintUnmergeable):
    def __init__(self, lst, operator):
        super().__init__(TypeCtr.LEX)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True, lifted=True)
        self.arg(TypeCtrArg.OPERATOR, operator)


class ConstraintLexMatrix(ConstraintUnmergeable):
    def __init__(self, lst, operator):
        super().__init__(TypeCtr.LEX)
        self.arg(TypeCtrArg.MATRIX, matrix_to_string(lst), content_compressible=lst)
        self.arg(TypeCtrArg.OPERATOR, operator)


''' Counting and Summing Constraints '''


class ConstraintSum(ConstraintWithCondition):
    def __init__(self, lst, coefficients, condition):
        super().__init__(TypeCtr.SUM)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=True)
        if coefficients is not None and any(not isinstance(v, int) or v != 1 for v in coefficients):
            self.arg(TypeCtrArg.COEFFS, coefficients, content_ordered=True)
        self.arg(TypeCtrArg.CONDITION, condition)

    def min_possible_value(self):
        vs = self.arguments[TypeCtrArg.LIST].content
        cs = self.arguments[TypeCtrArg.COEFFS].content if TypeCtrArg.COEFFS in self.arguments else None
        if cs is None:
            return sum(x.dom.smallest_value() for x in vs)
        assert len(vs) == len(cs)
        t = []
        for i, x in enumerate(vs):
            xmin, xmax = x.dom.smallest_value(), x.dom.greatest_value()
            if isinstance(cs[i], int):
                t.append(min(xmin * cs[i], xmax * cs[i]))
            else:
                cmin, cmax = cs[i].dom.smallest_value(), cs[i].dom.greatest_value()
                t.append(min(xmin * cmin, xmin * cmax, xmax * cmin, xmax * cmax))
        return sum(t)

    def max_possible_value(self):
        vs = self.arguments[TypeCtrArg.LIST].content
        cs = self.arguments[TypeCtrArg.COEFFS].content if TypeCtrArg.COEFFS in self.arguments else None
        if cs is None:
            return sum(x.dom.greatest_value() for x in vs)
        assert len(vs) == len(cs)
        t = []
        for i, x in enumerate(vs):
            xmin, xmax = x.dom.smallest_value(), x.dom.greatest_value()
            if isinstance(cs[i], int):
                t.append(max(xmin * cs[i], xmax * cs[i]))
            else:
                cmin, cmax = cs[i].dom.smallest_value(), cs[i].dom.greatest_value()
                t.append(max(xmin * cmin, xmin * cmax, xmax * cmin, xmax * cmax))
        return sum(t)


class ConstraintCount(ConstraintWithCondition):
    def __init__(self, lst, values, condition):
        super().__init__(TypeCtr.COUNT)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=False)
        self.arg(TypeCtrArg.VALUES, values)
        self.arg(TypeCtrArg.CONDITION, condition)

    def min_possible_value(self):
        return 0

    def max_possible_value(self):
        return len(self.arguments[TypeCtrArg.LIST].content)


class ConstraintNValues(ConstraintWithCondition):
    def __init__(self, lst, excepting, condition):
        super().__init__(TypeCtr.N_VALUES)
        self.arg(TypeCtrArg.LIST, lst)
        self.arg(TypeCtrArg.EXCEPT, excepting)
        self.arg(TypeCtrArg.CONDITION, condition)

    def min_possible_value(self):
        return 1

    def max_possible_value(self):
        return len(self.arguments[TypeCtrArg.LIST].content)


class ConstraintCardinality(Constraint):
    def __init__(self, lst, values, occurs, closed=False):
        super().__init__(TypeCtr.CARDINALITY)
        self.arg(TypeCtrArg.LIST, lst)
        assert len(values) == len(occurs)
        self.arg(TypeCtrArg.VALUES, values, content_ordered=True, attributes=[(TypeXML.CLOSED, "true")] if closed else [])
        self.arg(TypeCtrArg.OCCURS, occurs, content_ordered=True)


''' Connection Constraints '''


def _index_att(v):
    return [(TypeCtrArg.START_INDEX, v)] if v is not None and v != 0 else []


class ConstraintMaximum(ConstraintWithCondition):
    def __init__(self, lst, index=None, start_index=0, type_rank=TypeRank.ANY, condition=None):
        super().__init__(TypeCtr.MAXIMUM)
        self.arg(TypeCtrArg.LIST, lst, content_ordered=index is not None, attributes=_index_att(start_index))
        if index:
            self.arg(TypeCtrArg.INDEX, index, attributes=[(TypeCtrArg.RANK, type_rank)] if type_rank else [])
        self.arg(TypeCtrArg.CONDITION, condition)

    def min_possible_value(self):
        return max(x.dom.smallest_value() for x in self.arguments[TypeCtrArg.LIST].content)

    def max_possible_value(self):
        return max(x.dom.greatest_value() for x in self.arguments[TypeCtrArg.LIST].content)


class ConstraintMinimum(ConstraintMaximum):
    def __init__(self, lst, index=None, start_index=0, type_rank=TypeRank.ANY, condition=None):
        super().__init__(lst, index, start_index, type_rank, condition)
        self.name = TypeCtr.MINIMUM

    def min_possible_value(self):
        return min(x.dom.smallest_value() for x in self.arguments[TypeCtrArg.LIST].content)

    def max_possible_value(self):
        return min(x.dom.greatest_value() for x in self.arguments[TypeCtrArg.LIST].content)


class ConstraintElement(ConstraintWithCondition):  # currently, not exactly with a general condition
    def __init__(self, lst, index, value=None, type_rank=None):
        super().__init__(TypeCtr.ELEMENT)
        smallest = index.dom[0].smallest() if isinstance(index.dom[0], IntegerEntity) else index.dom[0]
        self.arg(TypeCtrArg.LIST, lst, content_ordered=index is not None, attributes=_index_att(smallest))
        self.arg(TypeCtrArg.INDEX, index, attributes=[(TypeCtrArg.RANK, type_rank)] if type_rank else [])
        self.arg(TypeCtrArg.VALUE, value)

    def min_possible_value(self):
        if isinstance(self.arguments[TypeCtrArg.LIST].content[0], int):
            return min(v for v in self.arguments[TypeCtrArg.LIST].content)
        return min(x.dom.smallest_value() for x in self.arguments[TypeCtrArg.LIST].content)

    def max_possible_value(self):
        if isinstance(self.arguments[TypeCtrArg.LIST].content[0], int):
            return max(v for v in self.arguments[TypeCtrArg.LIST].content)
        return max(x.dom.greatest_value() for x in self.arguments[TypeCtrArg.LIST].content)


class ConstraintElementMatrix(Constraint):
    def __init__(self, lst, index1, index2, value=None, start_row_index=0, start_col_index=0):
        super().__init__(TypeCtr.ELEMENT)
        self.arg(TypeCtrArg.MATRIX, matrix_to_string(lst), content_compressible=lst,
                 attributes=([(TypeCtrArg.ST.START_ROW_INDEX, start_row_index)] if start_row_index else []) + (
                     [(TypeCtrArg.START_COL_INDEX, start_col_index)] if start_col_index else []))
        self.arg(TypeCtrArg.INDEX, [index1, index2], content_ordered=True)
        self.arg(TypeCtrArg.VALUE, value)


class ConstraintChannel(ConstraintUnmergeable):
    def __init__(self, list1, start_index1, list2, start_index2):
        super().__init__(TypeCtr.CHANNEL)
        assert list1 is not None
        if list2 is None:  # first variant of channel
            assert start_index2 == 0
            self.arg(TypeCtrArg.LIST, list1, content_ordered=True, attributes=_index_att(start_index1))
        elif len(list2) > 1:  # second variant of channel
            assert 1 < len(list1) <= len(list2)
            self.arg(TypeCtrArg.LIST, [list1, list2], content_ordered=True, lifted=True, attributes=_index_att(start_index1) + _index_att(start_index2))
        else:  # third variant of channel
            assert all(var.dom.is_binary() for var in list1) and len(list2) == 1
            self.arg(TypeCtrArg.LIST, list1, content_ordered=True, attributes=_index_att(start_index1))
            self.arg(TypeCtrArg.VALUE, list2)


''' Packing and Scheduling Constraints '''


class ConstraintNoOverlap(ConstraintUnmergeable):
    def __init__(self, origins, lengths, zero_ignored):
        super().__init__(TypeCtr.NO_OVERLAP)
        self.arg(TypeCtrArg.ORIGINS, "".join(["(" + ",".join(str(v) for v in t) + ")" for t in origins]) if isinstance(origins[0], tuple) else origins,
                 content_ordered=True)
        self.arg(TypeCtrArg.LENGTHS, "".join(["(" + ",".join(str(v) for v in t) + ")" for t in lengths]) if isinstance(lengths[0], tuple) else lengths,
                 content_ordered=True)
        if zero_ignored:
            self.attributes.append((TypeCtrArg.ZERO_IGNORED, zero_ignored))


class ConstraintCumulative(Constraint):
    def __init__(self, origins, lengths, ends, heights, condition):
        super().__init__(TypeCtr.CUMULATIVE)
        self.arg(TypeCtrArg.ORIGINS, origins, content_ordered=True)
        self.arg(TypeCtrArg.LENGTHS, lengths, content_ordered=True)
        self.arg(TypeCtrArg.ENDS, ends, content_ordered=True)
        self.arg(TypeCtrArg.HEIGHTS, heights, content_ordered=True)
        self.arg(TypeCtrArg.CONDITION, condition)


''' Constraints on Graphs'''


class ConstraintCircuit(Constraint):
    def __init__(self, variables, start_index, size):
        super().__init__(TypeCtr.CIRCUIT)
        if start_index != 0:
            self.attributes.append((TypeCtrArg.START_INDEX, start_index))
        self.arg(TypeCtrArg.LIST, variables, content_ordered=True)
        self.arg(TypeCtrArg.SIZE, size)


''' Elementary Constraints '''


class ConstraintClause(Constraint):
    def __init__(self, variables, phases):
        super().__init__(TypeCtr.CLAUSE)
        self.arg(TypeCtrArg.LIST, [str(v) if phases[i] else "not(" + str(v) + ")" for i, v in enumerate(variables)], content_ordered=True)


class ConstraintInstantiation(Constraint):
    def __init__(self, variables, values):
        super().__init__(TypeCtr.INSTANTIATION)
        self.arg(TypeCtrArg.LIST, variables, content_ordered=True)
        self.arg(TypeCtrArg.VALUES, values, content_ordered=True)


''' PartialConstraints and ScalarProduct '''


class PartialConstraint:  # constraint whose condition has not been given such as 'Sum', 'Count', 'Element', 'maximum', etc.
    def __init__(self, constraint):
        self.constraint = constraint

    def add_condition(self, operator, right_operand):
        if isinstance(right_operand, (int, Variable)):
            return ECtr(self.constraint.set_condition(operator, right_operand))
        # TODO : which kind of right operand is authorized? just a partial sum?
        assert isinstance(self.constraint, ConstraintSum)
        pc = PartialConstraint.combine_partial_objects(self, TypeNode.SUB, right_operand)  # the 'complex' right operand is moved to the left
        return ECtr(pc.constraint.set_condition(operator, 0))

    def _simplify_with_auxiliary_variables(self, other):
        if isinstance(other, (int, Variable)):
            return other
        if isinstance(other, Node):
            return auxiliary().replace_node(other)
        assert isinstance(other, (ScalarProduct, PartialConstraint))
        if isinstance(self.constraint, ConstraintSum) and (isinstance(other, ScalarProduct) or isinstance(other.constraint, ConstraintSum)):
            return other
        assert isinstance(self.constraint, ConstraintWithCondition) and isinstance(other.constraint, ConstraintWithCondition)
        return auxiliary().replace_partial_constraint(other)

    def _simplify_operation(self, other):
        assert isinstance(self.constraint, ConstraintWithCondition)
        if isinstance(self.constraint, ConstraintSum) and (not isinstance(other, PartialConstraint) or isinstance(other.constraint, ConstraintSum)):
            return None  # we can combine partial sums and terms
        if not isinstance(self.constraint, ConstraintSum) and not isinstance(other, PartialConstraint):
            return auxiliary().replace_partial_constraint(self), other
        assert isinstance(other.constraint, ConstraintWithCondition)
        return auxiliary().replace_partial_constraint(self), auxiliary().replace_partial_constraint(other)

    def __eq__(self, other):
        other = self._simplify_with_auxiliary_variables(other)
        if isinstance(self.constraint, (ConstraintElement, ConstraintElementMatrix)) and isinstance(other, (int, Variable)):
            if isinstance(self.constraint, ConstraintElement):
                arg = self.constraint.arguments[TypeCtrArg.LIST]
                arg.content = flatten(arg.content)  # we need to flatten now because it has not been done before
            return ECtr(self.constraint.set_value(other))  # only value must be replaced for these constraints
        return self.add_condition(TypeConditionOperator.EQ, other)
        # pair = self._simplify_with_auxiliary_variables(other)
        # return Node.build(TypeNode.EQ, pair) if pair else self.add_condition(TypeConditionOperator.EQ, other)

    def __ne__(self, other):
        return self.add_condition(TypeConditionOperator.NE, self._simplify_with_auxiliary_variables(other))

    def __lt__(self, other):
        return self.add_condition(TypeConditionOperator.LT, self._simplify_with_auxiliary_variables(other))

    def __le__(self, other):
        return self.add_condition(TypeConditionOperator.LE, self._simplify_with_auxiliary_variables(other))

    def __ge__(self, other):
        return self.add_condition(TypeConditionOperator.GE, self._simplify_with_auxiliary_variables(other))

    def __gt__(self, other):
        return self.add_condition(TypeConditionOperator.GT, self._simplify_with_auxiliary_variables(other))

    def __add__(self, other):
        pair = self._simplify_operation(other)
        return Node.build(TypeNode.ADD, pair) if pair else PartialConstraint.combine_partial_objects(self, TypeNode.ADD, other)

    def __sub__(self, other):
        pair = self._simplify_operation(other)
        return Node.build(TypeNode.SUB, pair) if pair else PartialConstraint.combine_partial_objects(self, TypeNode.SUB, other)

    def __mul__(self, other):
        assert isinstance(self.constraint, ConstraintSum) and isinstance(other, int)
        args = self.constraint.arguments
        cs = args[TypeCtrArg.COEFFS].content if TypeCtrArg.COEFFS in args else [1] * len(args[TypeCtrArg.LIST].content)
        value = args[TypeCtrArg.CONDITION]
        del args[TypeCtrArg.CONDITION]  # we delete and put back below this argument so as to have arguments in the right order
        self.constraint.arg(TypeCtrArg.COEFFS, [c * other for c in cs])
        args[TypeCtrArg.CONDITION] = value
        return self

    def __rmul__(self, other):
        return PartialConstraint.__mul__(self, other)

    def __getitem__(self, i):
        assert isinstance(self.constraint, ConstraintElement)
        lst = self.constraint.arguments[TypeCtrArg.LIST].content
        index = self.constraint.arguments[TypeCtrArg.INDEX].content
        if isinstance(i, Variable):
            assert is_matrix(lst), "Variables in element constraint must be in the form of matrix"
            self.constraint = ConstraintElementMatrix(lst, index, i)
        elif isinstance(i, int):
            self.constraint = ConstraintElement(lst[:, i], index)
        return self

    def __str__(self):
        c = self.constraint
        # assert len(c.arguments) == 2
        return str(c.name) + "(" + str(compact(c.arguments[TypeCtrArg.LIST].content)) + ")"  # TODO experimental stuff

    @staticmethod
    def combine_partial_objects(obj1, operator, obj2):  # currently, only partial sums can be combined
        assert operator in {TypeNode.ADD, TypeNode.SUB}
        if isinstance(obj1, ScalarProduct):
            obj1 = PartialConstraint(ConstraintSum(obj1.variables, obj1.coeffs, None))  # to be sure to have at least one PartialConstraint
        assert isinstance(obj1, PartialConstraint) or isinstance(obj2, PartialConstraint)
        inverted, obj1, obj2 = (False, obj1, obj2) if isinstance(obj1, PartialConstraint) else (True, obj2, obj1)
        pair = obj2.var_val_if_binary_type(TypeNode.MUL) if isinstance(obj2, Node) else None
        if pair:
            obj2 = PartialConstraint(ConstraintSum([pair[0]], [pair[1]], None))
        elif isinstance(obj2, Variable):
            obj2 = PartialConstraint(ConstraintSum([obj2], [1], None))
        elif isinstance(obj2, ScalarProduct):
            obj2 = PartialConstraint(ConstraintSum(obj2.variables, obj2.coeffs, None))
        elif not isinstance(obj2, PartialConstraint):
            error("The type of the operand of the partial constraint is wrong as it is " + str(type(obj2)))
        obj1, obj2 = (obj1, obj2) if not inverted else (obj2, obj1)  # we invert back
        assert isinstance(obj1, PartialConstraint) and isinstance(obj2, PartialConstraint)
        assert isinstance(obj1.constraint, ConstraintSum) and isinstance(obj2.constraint, ConstraintSum)
        args1, args2 = obj1.constraint.arguments, obj2.constraint.arguments
        vs1, vs2 = args1[TypeCtrArg.LIST].content, args2[TypeCtrArg.LIST].content
        cs1 = args1[TypeCtrArg.COEFFS].content if TypeCtrArg.COEFFS in args1 else [1] * len(vs1)
        cs2 = args2[TypeCtrArg.COEFFS].content if TypeCtrArg.COEFFS in args2 else [1] * len(vs2)
        cs2 = [-c for c in cs2] if operator == TypeNode.SUB else cs2
        return PartialConstraint(ConstraintSum(vs1 + vs2, cs1 + cs2, None))


class ScalarProduct:
    def __init__(self, variables, coefficients):
        variables = list(variables) if isinstance(variables, tuple) else variables
        coefficients = list(coefficients) if isinstance(coefficients, tuple) else coefficients
        assert isinstance(variables, list) and isinstance(coefficients, (int, list, range)), variables
        self.variables = flatten(variables)  # for example, in order to remove None occurrences
        self.coeffs = flatten([coefficients] * len(variables) if isinstance(coefficients, int) else coefficients)
        assert len(self.variables) == len(self.coeffs), str(self.variables) + " " + str(self.coeffs)

    def _combine_with(self, operator, right_operand):
        return PartialConstraint(ConstraintSum(self.variables, self.coeffs, None)).add_condition(operator, right_operand)

    def __lt__(self, other):
        return self._combine_with(TypeConditionOperator.LT, other)

    def __le__(self, other):
        return self._combine_with(TypeConditionOperator.LE, other)

    def __ge__(self, other):
        return self._combine_with(TypeConditionOperator.GE, other)

    def __gt__(self, other):
        return self._combine_with(TypeConditionOperator.GT, other)

    def __eq__(self, other):
        return self._combine_with(TypeConditionOperator.EQ, other)

    def __ne__(self, other):
        return self._combine_with(TypeConditionOperator.NE, other)

    def __add__(self, other):
        return PartialConstraint.combine_partial_objects(self, TypeNode.ADD, other)

    def __sub__(self, other):
        return PartialConstraint.combine_partial_objects(self, TypeNode.SUB, other)


class _Auxiliary:
    def __init__(self):
        self._introduced_variables = []
        self._collected_constraints = []
        self.prefix = "aux_gb"

    def __replace(self, obj, dom):
        assert dom.get_type() == TypeVar.INTEGER
        index = len(self._introduced_variables)
        name = self.prefix + "[" + str(index) + "]"
        var = VariableInteger(name, dom)
        Variable.name2obj[name] = var
        if index == 0:
            self._introduced_variables = EVarArray([var], self.prefix, self.prefix + "[i] is the ith auxiliary variable having been automatically introduced")
        else:
            self._introduced_variables.extend_with(var)
        self._collected_constraints.append((obj, var))
        return var

    def replace_partial_constraint(self, pc):
        assert isinstance(pc, PartialConstraint)
        return self.__replace(pc, Domain(range(pc.constraint.min_possible_value(), pc.constraint.max_possible_value() + 1)))

    def replace_partial_constraints(self, terms):
        assert isinstance(terms, list)
        for i, t in enumerate(terms):
            if isinstance(t, PartialConstraint):
                terms[i] = self.replace_partial_constraint(t)
        return terms

    def replace_node(self, node):
        assert isinstance(node, Node)
        values = node.possible_values()
        values = {v for v in values} if len(values) <= 2 else values  # in order to avoid having a range for just 1 or 2 values
        return self.__replace(node, Domain(values))

    def collected(self):
        t = self._collected_constraints
        self._collected_constraints = []
        return t


def auxiliary():
    if not hasattr(auxiliary, "obj"):
        auxiliary.obj = _Auxiliary()
    return auxiliary.obj



"""
See https://en.wikipedia.org/wiki/Graeco-Latin_square

A Latin square of order n is an n by n array filled with n different symbols (for example, values between 1 and n),
each occurring exactly once in each row and exactly once in each column.
Two latin squares of the same order n are orthogonal if each pair of elements in the same position occurs exactly once.
The most easy way to see this is by concatenating elements in the same position and verify that no pair appears twice.
There are orthogonal latin squares of any size except 1, 2, and 6.

Examples of Execution:
  python3 OrthoLatin.py
  python3 OrthoLatin.py -data=7
"""

from pycsp3 import *

n = data or 5

# x is the first Latin square
x = VarArray(size=[n, n], dom=range(n))

# y is the second Latin square
y = VarArray(size=[n, n], dom=range(n))

# z is the matrix used to control orthogonality
z = VarArray(size=[n * n], dom=range(n * n))

table = {(i, j, i * n + j) for i in range(n) for j in range(n)}

satisfy(
    # ensuring that x is a Latin square
    AllDifferent(x, matrix=True),

    # ensuring that y is a Latin square
    AllDifferent(y, matrix=True),

    # ensuring that values on diagonals are different  tag(diagonals)
    [AllDifferent(t) for t in [diagonal_down(x), diagonal_up(x), diagonal_down(y), diagonal_up(y)]],

    # ensuring orthogonality of x and y through z
    AllDifferent(z),

    # computing z from x and y
    [(x[i][j], y[i][j], z[i * n + j]) in table for i in range(n) for j in range(n)],

    # tag(symmetry-breaking)
    [(x[0][j] == j, y[0][j] == j) for j in range(n)]
)


# Note that a less compact way of posting symmetry_breaking constraints is:
# # tag(symmetry-breaking)
# [[x[0][j] == j for j in range(n)], [y[0][j] == j for j in range(n)]],



"""
Well-known crypto-arithmetic puzzle of unknown origin (e.g., a model is present in Gecode)

Examples of Execution:
  python3 Alpha.py
  python3 Alpha.py -variant=var
"""

from pycsp3 import *

if not variant():
    def of(word):
        return [x[i] for i in alphabet_positions(word)]


    # x[i] is the value for the ith letter of the alphabet
    x = VarArray(size=26, dom=range(1, 27))

    satisfy(
        # all letters must be different
        AllDifferent(x),

        # respecting clues
        [Sum(of("ballet")) == 45,
         Sum(of("cello")) == 43,
         Sum(of("concert")) == 74,
         Sum(of("flute")) == 30,
         Sum(of("fugue")) == 50,
         Sum(of("glee")) == 66,
         Sum(of("jazz")) == 58,
         Sum(of("lyre")) == 47,
         Sum(of("oboe")) == 53,
         Sum(of("opera")) == 65,
         Sum(of("polka")) == 59,
         Sum(of("quartet")) == 50,
         Sum(of("saxophone")) == 134,
         Sum(of("scale")) == 51,
         Sum(of("solo")) == 37,
         Sum(of("song")) == 61,
         Sum(of("soprano")) == 82,
         Sum(of("theme")) == 72,
         Sum(of("violin")) == 100,
         Sum(of("waltz")) == 34]
    )

elif variant("var"):
    # letters[i] is the value for the ith letter of the alphabet
    a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z = letters = VarArray(size=26, dom=range(1, 27))

    satisfy(
        # all letters must be different
        AllDifferent(letters),

        # respecting clues
        [Sum(b, a, l, l, e, t) == 45,
         Sum(c, e, l, l, o) == 43,
         Sum(c, o, n, c, e, r, t) == 74,
         Sum(f, l, u, t, e) == 30,
         Sum(f, u, g, u, e) == 50,
         Sum(g, l, e, e) == 66,
         Sum(j, a, z, z) == 58,
         Sum(l, y, r, e) == 47,
         Sum(o, b, o, e) == 53,
         Sum(o, p, e, r, a) == 65,
         Sum(p, o, l, k, a) == 59,
         Sum(q, u, a, r, t, e, t) == 50,
         Sum(s, a, x, o, p, h, o, n, e) == 134,
         Sum(s, c, a, l, e) == 51,
         Sum(s, o, l, o) == 37,
         Sum(s, o, n, g) == 61,
         Sum(s, o, p, r, a, n, o) == 82,
         Sum(t, h, e, m, e) == 72,
         Sum(v, i, o, l, i, n) == 100,
         Sum(w, a, l, t, z) == 34]
    )



"""
See Problem 034 on CSPLib

Examples of Execution:
  python3 Warehouse.py -data=Warehouse_example.json
  python3 Warehouse.py -data=Warehouse_example.txt -dataparser=Warehouse_Parser.py
"""

from pycsp3 import *

cost, capacities, costs = data  # cost is the fixed cost when opening a warehouse
nWarehouses, nStores = len(capacities), len(costs)

# w[i] is the warehouse supplying the ith store
w = VarArray(size=nStores, dom=range(nWarehouses))

# c[i] is the cost of supplying the ith store
c = VarArray(size=nStores, dom=lambda i: costs[i])

# o[j] is 1 if the jth warehouse is open
o = VarArray(size=nWarehouses, dom={0, 1})

satisfy(
    # capacities of warehouses must not be exceeded
    [Count(w, value=j) <= capacities[j] for j in range(nWarehouses)],

    # the warehouse supplier of the ith store must be open
    [o[w[i]] == 1 for i in range(nStores)],

    # computing the cost of supplying the ith store
    [costs[i][w[i]] == c[i] for i in range(nStores)]
)

minimize(
    # minimizing the overall cost
    Sum(c) + Sum(o) * cost
)



"""
See, e.g., "Solving the asymmetric traveling purchaser problem" by J. Riera-Ledesma, J. Salazar González, Annals OR 144(1): 83-97 (2006)
See similar model (called ttp) proposed by Kathryn Francis for the 2012 Minizinc Competition

Examples of Execution:
  python3 TravelingPurchaser.py -data=TravelingPurchaser-7-5-30-1.json
"""

from pycsp3 import *

distances, prices = data
nCities, nProducts = len(distances), len(prices)

# s[i] is the city succeeding to the ith city (itself if not part of the route)
s = VarArray(size=nCities, dom=range(nCities))

# d[i] is the distance (seen as a travel cost) between the ith city and its successor
d = VarArray(size=nCities, dom=lambda i: {v for v in distances[i] if v >= 0})

# l[i] is the purchase location of the ith product (last city has nothing for sale)
l = VarArray(size=nProducts, dom=range(nCities - 1))

# c[i] is the purchase cost of the ith product
c = VarArray(size=nProducts, dom=lambda i: set(prices[i]))

satisfy(
    # linking distances to successors
    [distances[i][s[i]] == d[i] for i in range(nCities)],

    # linking purchase locations to purchase costs 
    [prices[i][l[i]] == c[i] for i in range(nProducts)],

    # purchasing a product at a city is only possible if you visit that city 
    [imply(s[i] == i, l[j] != i) for i in range(nCities) for j in range(nProducts)],

    Circuit(s),

    # last city must be visited (we start here)
    s[nCities - 1] != nCities - 1
)

minimize(
    Sum(d) + Sum(c)
)



#  sudo pip3 install py4j && sudo pip3 install pandas && sudo pip3 install xlrd

import datetime
import json
import math
from collections import OrderedDict

from pycsp3 import *
from pycsp3.solvers.abscon import AbsConProcess

date = "04/09/2019"  # hard coding for the moment (see also in the creation of the JSON solution file)
k_arrival, k_departure = 300, 300  #  hard coding for the time security in seconds

# note that data.ordonnancement, data.reductions, data.traitement and data.vols are not used

# ### about parkings
parkings = data.parkings['Code']
parking_indexes = OrderedDict((parking, i) for i, parking in enumerate(parkings))  # keys are parkings (actually, their codes) and values their indexes
possible_parkings = OrderedDict()  # per plane type
for i, ta in enumerate(data.capacites['TA_CLE']):
    possible_parkings.setdefault(ta, []).append(data.capacites['RSC_COD'][i])
nParkings = len(parkings)

# ### about flights
flights = [{key: data.volsStrategies[key][i] for key in data.volsStrategies.keys()} for i in range(len(data.volsStrategies['Date arr']))
           if date in {data.volsStrategies['Date arr'][i].split(" ")[0], data.volsStrategies['Date dep'][i].split(" ")[0]}]
nFlights = len(flights)


def table_capacities(flight):
    return {parking_indexes[parking] for parking in possible_parkings[flight['Type Avion']]}


def table_rewards(flight):
    target_parkings = [strategy.split("/")[1] for strategy in data.strategies['Ressource'] if "/" in strategy]  # parkings without the character "/"
    table = set()
    for i, parking in enumerate(parkings):
        value = data.strategies[flight['Stratégie code dep']][target_parkings.index(parking)] if parking in target_parkings else 0
        table.add((i, int(value) if not math.isnan(value) else 0))
    return table


def table_shadings():
    p1, p2 = data.ombrages['RSC_COD'], data.ombrages['RSC_COD_OMBRE']
    assert len(p1) == len(p2)
    return (
        {(i, i) for i in range(nParkings)}  # conflict if same parking
        | {(parking_indexes[p1[i]], parking_indexes[p2[i]]) for i in range(len(p1))}  # conflict if shading between parkings
    )


def to_datetime(s):
    part1, part2 = s.split(" ")
    day, month, year, hour, minute, second = [int(v) for v in part1.split("/")] + [int(v) for v in part2.split(":")]
    return datetime.datetime(year, month, day, hour, minute, second)


def are_overlapping(flight1, flight2):
    arr1 = to_datetime(flight1['Date arr']) - datetime.timedelta(seconds=k_arrival)
    dep1 = to_datetime(flight1['Date dep']) + datetime.timedelta(seconds=k_departure)
    arr2 = to_datetime(flight2['Date arr']) - datetime.timedelta(seconds=k_arrival)
    dep2 = to_datetime(flight2['Date dep']) + datetime.timedelta(seconds=k_departure)
    return arr1 < dep2 and arr2 < dep1


# p[i] is the parking (code) of the ith flight
p = VarArray(size=nFlights, dom=range(nParkings))

# r[i] is the reward (strategy satisfaction between 0 and 100) of the ith flight  
r = VarArray(size=nFlights, dom=range(101))

satisfy(
    p[0] == 2,
    # taking into account only parkings authorized for flights
    [p[i] in table_capacities(flight) for i, flight in enumerate(flights)],

    # computing rewards
    [(p[i], r[i]) in table_rewards(flight) for i, flight in enumerate(flights)],

    # taking into account shadings
    [(p[i], p[j]) not in table_shadings() for i, j in combinations(range(nFlights), 2) if are_overlapping(flights[i], flights[j])]
)

maximize(
    r * [flight['Poids total rotation'] for flight in flights]
)

annotate(
    decision=r
)

###
# Below, compilation and execution of the solver
###

instance = compile()
solution = AbsConProcess().solve(instance, n_restarts=3300)  # 1000)
print("\n",solution)

if solution and solution.variables:
    def pretty_flight(i, flight):
        plane, company = flight['Type Avion'], flight['Comapgnie arr']
        arrival = date + " 00:00:00" if flight['Date arr'][0:2] == "03" else flight['Date arr']  # hard coding for 03
        departure = date + " 23:55:00" if flight['Date dep'][0:2] == "05" else flight['Date dep']  # hard coding for 05
        parking, reward = parkings[int(solution.values[i])], int(solution.values[nFlights + i])
        n5 = int((to_datetime(departure) - to_datetime(arrival)).total_seconds()) / 300  # number of slots of 5 minutes
        return {"index": i, "plane": plane, "company": company, "arrival": arrival, "departure": departure, "parking": parking, "reward": reward, "n5": n5}

    with open("g7_todo/solutionPlaneParking.json", 'w') as g:
        g.write("let flights = ")
        json.dump([pretty_flight(i, flight) for i, flight in enumerate(flights)], g, separators=(',', ':'))
    print("Generation of the JSON solution file solutionPlaneParking.json completed.")


# solver = AbsconPy4J()
# solver.loadXCSP3(xml)



"""
See https://mathworld.wolfram.com/TalismanSquare.html

An n x n array  of the integers from 1 to n^2 such that the difference between any integer
and its neighbor (horizontally, vertically, or diagonally, without wrapping around)
is greater than or equal to some value k is called a (n,k)-talisman square.

Examples of Execution:
  python3 Talisman.py
  python3 Talisman.py -data=[5,4]
"""

from pycsp3 import *

n, k = data or (4, 2)
limit = (n * (n * n + 1)) // 2

# x[i][j] is the value in the talisman square at row i and column j
x = VarArray(size=[n, n], dom=range(1, n * n + 1))

satisfy(
    # all values must be different
    AllDifferent(x),

    # the distance between two neighbouring cells must be strictly greater than k
    [
        [abs(x[i][j] - x[i][j + 1]) > k for i in range(n) for j in range(n - 1)],
        [abs(x[i][j] - x[i + 1][j]) > k for j in range(n) for i in range(n - 1)],
        [abs(dgn[i] - dgn[i + 1]) > k for dgn in diagonals_down(x) for i in range(len(dgn) - 1)],
        [abs(dgn[i] - dgn[i + 1]) > k for dgn in diagonals_up(x) for i in range(len(dgn) - 1)]
    ],

    # tag(symmetry-breaking)
    x[0][0] == 1
)



"""
Problem 018 on CSPLib

Examples of Execution:
  python3 WaterBucket.py
  python3 WaterBucket.py -data=[8,5,3,4,4,0,8]
"""

from pycsp3 import *

# ci for capacities of the three buckets
# gi for goal (which quantities must be present in the three buckets after all transfers)
# h for horizon (maximal number of rounds/transfers)
c1, c2, c3, g1, g2, g3, h = data or (8, 5, 3, 4, 4, 0, 8)

assert c1 >= c2 >= c3 > 0, "Bucket capacities must be in decreasing order"
assert g1 + g2 + g3 == c1, "water from bucket 1 must be split into the three buckets to reach the goal"


def tables():
    def nearby_states(a1, a2, a3, b1, b2, b3):
        if (a1, a2, a3) == (g1, g2, g3):
            return (b1, b2, b3) == (g1, g2, g3)  # when the goal is reached, buckets are kept as they are
        if a1 != b1 and a2 != b2 and a3 != b3:  # not a possible transition if three buckets are impacted
            return False
        if a1 != b1 and a2 != b2:
            return b1 in {0, c1} or b2 in {0, c2}
        if a1 != b1 and a3 != b3:
            return b1 in {0, c1} or b3 in {0, c3}
        if a2 != b2 and a3 != b3:
            return b2 in {0, c2} or b3 in {0, c3}
        return False  # not related in other cases

    states = {(i, j, k) for i in range(c1 + 1) for j in range(c2 + 1) for k in range(c3 + 1) if i + j + k == c1}
    table_tr = {(*state1, *state2) for state1 in states for state2 in states if nearby_states(*state1, *state2)}
    table_hr = {(t,) + (ANY,) * (t * 3) + (g1, g2, g3) * (h - t) for t in range(h)}
    return table_tr, table_hr


table_transitions, table_horizon = tables()

# x[t][i] is the volume of water in bucket i at time (round) t
x = VarArray(size=[h, 3], dom=[[range(c1 + 1), range(c2 + 1), range(c3 + 1)]] * h)

# z is the number of transfers of water in order to reach the goal
z = Var(dom=range(h))

satisfy(
    # Initially, at round 0, the bucket 1 is full of water while the other buckets are empty
    x[0] == (c1, 0, 0),

    # only some transfers are possible between two successive rounds
    [(x[t] + x[t + 1]) in table_transitions for t in range(h - 1)],  # or (*x[t], *x[t + 1])

    # computing the number of transfers
    (z, *flatten(x)) in table_horizon
)

minimize(
    # minimizing the number of transfers
    z
)



"""
Problem 110 on CSPLib

Examples of Execution:
  python3 PeacableArmies.py -data=10 -variant=m1
  python3 PeacableArmies.py -data=10 -variant=m2
"""

from pycsp3 import *

n = data or 6

if variant("m1"):

    def less_equal(i1, j1, i2, j2):
        if (i1, j1) == (i2, j2):
            return b[i1][j1] + w[i1][j1] <= 1
        if i1 < i2 or (i1 == i2 and j1 < j2):
            if i1 == i2 or j1 == j2 or abs(i1 - i2) == abs(j1 - j2):  # maybe we can simplify something here
                return b[i1][j1] + w[i2][j2] <= 1, w[i1][j1] + b[i2][j2] <= 1


    # b[i][j] is 1 if a black queen is in the cell at row i and column j
    b = VarArray(size=[n, n], dom={0, 1})

    # w[i][j] is 1 if a white queen is in the cell at row i and column j
    w = VarArray(size=[n, n], dom={0, 1})

    satisfy(
        # no two opponent queens can attack each other
        [less_equal(i1, j1, i2, j2) for (i1, j1, i2, j2) in product(range(n), repeat=4)],

        # ensuring the same numbers of black and white queens
        Sum(b) == Sum(w)
    )

    maximize(
        # maximizing the number of black queens (and consequently, the size of the armies)
        Sum(b)
    )

if variant("m2"):

    def different(i1, j1, i2, j2):
        if i1 < i2 or (i1 == i2 and j1 < j2):
            if i1 == i2 or j1 == j2 or abs(i1 - i2) == abs(j1 - j2):
                return x[i1][j1] + x[i2][j2] != 3


    # x[i][j] is 1 (resp., 2), if a black (resp., white) queen is in the cell at row i and column j. It is 0 otherwise.
    x = VarArray(size=[n, n], dom={0, 1, 2})

    # nb is the number of black queens
    nb = Var(dom=range(n * n // 2))

    # nw is the number of white queens
    nw = Var(dom=range(n * n // 2))

    satisfy(
        # no two opponent queens can attack each other
        [different(i1, j1, i2, j2) for (i1, j1, i2, j2) in product(range(n), repeat=4)],

        # counting the number of black queens
        Count(x, value=1) == nb,

        # counting the number of white queens
        Count(x, value=2) == nw,

        # ensuring equal-sized armies
        nb == nw
    )

    maximize(
        # maximizing the number of black queens (and consequently, the size of the armies)
        nb
    )



import inspect
import os
import sys

from pycsp3.dashboard import options
from pycsp3.tools.utilities import flatten

if not os.name == 'nt':
    import readline

fileToTrace = [arg for arg in sys.argv if arg.endswith(".py") and "compiler.py" not in arg] + ["<stdin>"]


def _global_value_of(variable):
    stack = list(reversed(inspect.stack(context=1)))
    frames = [frame for frame in stack if frame.filename in fileToTrace]
    if len(frames) == 0:
        return None
    globs = frames[0].frame.f_globals
    return str(globs[variable]) if variable in globs.keys() else None


def is_comment_line(line):
    if not isinstance(line, str):
        return False
    line = line.strip()
    return len(line) > 2 and line[0] == '#' and line.replace(" ", "")[1] != '!'


def is_empty_line(line):
    return isinstance(line, str) and len(line.strip()) == 0


def is_continued_line(line):
    return isinstance(line, str) and line.strip().endswith('\\')


def is_correct_frame(frame, f):
    return frame.function == f or frame.function == "__init__" and f in frame.code_context[0]


def _extract_code(function_name):
    stack = list(reversed(inspect.stack(context=1)))
    frame = [(i - 1, stack[i - 1]) for i, frame in enumerate(stack) if is_correct_frame(frame, function_name) and i > 0][0]  # Get the correct frame
    if frame[1].filename == "<stdin>":  # Console mode
        if os.name == 'nt':
            assert os.name != 'nt', "Console mode is not available on Windows"
        lines = reversed([readline.get_history_item(i + 1) for i in range(readline.get_current_history_length())])
    else:  # File mode
        frame = list(reversed(inspect.stack(context=100)))[frame[0]]  # Get the same frame but with at more 100 line of codes
        lines = list(reversed(frame.code_context[:frame.index + 1]))
    # Now, get code attached to the function (empty lines, comment lines, lines continued with '\', and code lines of the function)
    code = []
    found = False  # the function name
    for line in lines:  # note that we process upward
        if found is False:
            code.append(line)
            if function_name in line and not is_comment_line(line):
                found = True
        else:
            if is_continued_line(line) or is_empty_line(line) or is_comment_line(line):
                code.append(line)
            else:
                break
    return code


# returns a pair (left,right) of positions of the first occurrence of the specified separators in the line, or None
def _find_separators(line, left_separator, right_separator):
    left = line.find(left_separator)
    if left == -1:
        return None
    right = line[left + len(left_separator):].find(right_separator)
    return (left, left + len(left_separator) + right) if right != -1 else None


def _find_tags(line):
    positions = _find_separators(line, "tag(", ")")
    return " ".join(line[positions[0] + 4: positions[1]].split(",")) if positions else None


# remove tags if any, and get global values (i.e, values of variables whose names are between quotes)
def _prepare(line):
    while True:
        positions = _find_separators(line, "'", "'")
        if positions is None:
            break
        left, right = positions
        value = _global_value_of(line[left + 1:right])
        if value:
            line = line[:left] + value + line[right + 1:]
        else:
            break  # we abandon the process
    positions = _find_separators(line, "tag(", ")")
    if positions:
        line = line[:positions[0]] + line[positions[1] + 1:]
    return None if line == "" or line.strip() == "#" else line.strip()[1:].strip()


def comment_and_tags_of(*, function_name):
    lines = []
    found = False  # the function name
    for line in _extract_code(function_name):  # code is given bottom-up
        if not is_comment_line(line) and function_name in line:
            found = True
            continue
        elif is_comment_line(line) and found:
            lines.append(line)
        elif is_empty_line and found:
            break
    if len(lines) == 0:
        return None, None
    comment, tags = "", ""
    for line in reversed(lines):
        if _find_tags(line) is not None:
            tags += ("" if len(tags) == 0 else " ") + _find_tags(line).strip()
        if _prepare(line) is not None:
            comment += ("" if len(comment) == 0 else " " if comment[-1] in {'.', ','} else " - ") + _prepare(line)
    return comment, tags


def _remove_matching_brackets(line):
    bracket = None
    for i, c in enumerate(line):
        if c in {'(', '[', '{'}:  # todo add 'for'?
            left = i
            bracket = c
        elif (c, bracket) in {(')', '('), (']', '['), ('}', '{')}:  # todo add ('for', 'in') ?
            right = i + 1
            break
    else:  # no break
        return line
    return _remove_matching_brackets(line[:left] + line[right:])  # recursive call


# Delete ( ) parts and [ ] parts of each line excepts for comment lines
def _delete_bracket_part(code, nb_parameters):
    # if nb_parameters == 1 and len(code) == 1:  # TODO  code for case like variant aux-v7 in Quasigroup (remonving ,)
    #     code = [line.strip() if is_comment_line(line) else line.replace(",", "").strip() for line in code]
    # else:
    code = [line.strip() if is_comment_line(line) else _remove_matching_brackets(line).strip() for line in code]
    if len(code) > 0 and len(code[-1]) > 0 and code[-1][-1] == ',':  # removing useless trailing comma of specify() if any
        code[-1] = code[:-1]
    return code


def comments_and_tags_of_parameters_of(*, function_name, args):
    if len(args) == 0:
        return [], [], [], []

    comments1 = [""] * len(args)  # comments at first level (i.e., at level of the parameters)
    comments2 = [[""] for _ in args]  # comments at second level (i.e, for elements of intern lists)
    tags1 = [""] * len(args)  # tags at first level
    tags2 = [[""] for _ in args]  # tags at second level
    code = list(reversed(_extract_code(function_name)))
    are_empty_lines = [is_empty_line(line) for line in code]

    code = _delete_bracket_part(code, len(args))

    level = 0
    i1 = 0  # index indicating the position of the current parameter
    i2 = 0  # index indicating the position of the element of the current list (inside the current parameter)
    found = False
    for i, line in enumerate(code):
        if not is_comment_line(line):
            new_line = ""  # possibly a new line to replace line
            for j, c in enumerate(line):
                if c == '#':  # if comment in code line
                    break
                if function_name in line:
                    if j < line.index(function_name):  # Fix for '(' or '[' before the name of the function
                        continue
                    else:
                        found = True  # Fix function name for comment of level 2
                if c in {'(', '['}:
                    level += 1
                elif c in {')', ']'}:
                    level -= 1
                elif c == ",":
                    if level == 1:
                        i1 += 1
                        i2 = 0
                        new_line += c
                    elif level == 2:
                        i2 += 1
                        comments2[i1].append("")
                        tags2[i1].append("")
                        new_line += '$'
                    else:
                        assert level != 0
                        new_line += '_'
                else:
                    new_line += c
            code[i] = "" if level != 1 else new_line

        if found and level == 2 and is_comment_line(line):
            s = _prepare(line)
            comments2[i1][i2] += ("" if len(comments2[i1][i2]) == 0 else " - ") + ("" if s is None else s)
            tags = _find_tags(line)
            if tags:
                tags2[i1][i2] += ("" if len(tags2[i1][i2]) == 0 else " ") + tags
            code[i] = ""

    # n_commas = sum(s.count(',') for s in code if not is_comment_line(s))  # TODO problem with subvariant aux-v7 of Quasigroup
    # assert len(args) == n_commas + 1, "Inspector error: numbers of commas incorrect in satisfy():" + str(n_commas) + " in " + str(code)

    # collecting comments and tags for each parameter (i.e., at level 1)
    i1 = 0
    found = False  # the function name
    for i, line in enumerate(code):
        if function_name in line:
            found = True
        if found and is_comment_line(line) and _prepare(line) and i + 1 < len(code):
            comments1[i1] += ("" if len(comments1[i1]) == 0 else " " if comments1[i1][-1] in {'.', ','} else " - ") + _prepare(line)
            if are_empty_lines[i + 1]:
                comments1[i1] = ""
        tags = _find_tags(line)
        if tags:
            tags1[i1] += ("" if len(tags1[i1]) == 0 else " ") + tags
        elif not is_comment_line(line) and ',' in line:
            i1 += line.count(',')
    return comments1, comments2, tags1, tags2


def extract_declaration_for(function_name):
    code = list(reversed(_extract_code(function_name)))
    for line in code:
        if function_name in line and not is_comment_line(line):
            pos = line.find(function_name)
            if "=" in line[:pos]:
                break
    else:
        assert False, " the object returned by " + function_name + " should be assigned to a variable"
    declaration = line[:pos].strip()
    if declaration[-1] == '=':
        declaration = declaration[:-1].strip()
    assert declaration.count('=') < 2
    if '=' in declaration:
        t = declaration.split('=')
        declaration = t[0] if ',' in t[0] else t[1]
    if function_name == "Var":
        assert "," not in declaration and ")" not in declaration, \
            "Every simple declaration must be on its own line. For example, 'x, y = Var(dom={0,1}), Var(dom={0,1})' is not allowed."
        return declaration
    elif function_name == "VarArray":
        assert ")" not in declaration
        return declaration if "," not in declaration else declaration.split(",")


# def docstringOf(name):
#    att = getattr(functions, name, None)
#    return None if att is None else att.__doc__


def checkType(obj, allowed_types, message=""):
    if options.checker == "none":
        return True
    if options.checker == "fast" and isinstance(obj, (list, tuple, set)) and len(obj) > 100:
        obj = obj[:1]
    allowed_types = (allowed_types,) if not isinstance(allowed_types, tuple) else allowed_types
    for allowedType in allowed_types:
        if not isinstance(allowedType, list):
            if isinstance(obj, allowedType):
                return True
        elif isinstance(obj, (list, tuple, set)):
            for p in flatten(obj):
                if not any(isinstance(p, typ) for typ in allowedType):
                    break  # raise TypeMCSPError(inspector.getCalling(), p, allowedTypes, message, position)
            else:
                return True
    # stack = inspect.stack(context=1)
    # name_function_stack = [s.function for s in stack]
    # position_functions_file = [i for i, s in enumerate(stack) if s.filename.endswith("functions.py")][0]
    # modeling_file = stack[position_functions_file + 2] if len(stack) > position_functions_file + 2 else None
    # if modeling_file is None and "<module>" in name_function_stack:
    #     modeling_file = stack[[i for i, s in enumerate(stack) if s.function == "<module>"][0]]
    # function_name = "satisfy"
    # problem_lines = []
    # if function_name in name_function_stack: #  satisfy case
    #     code = list(reversed(_extract_code(function_name)))
    #     if len(code) > 0:
    #         without_bracket_code = _delete_bracket_part(code, functions.nb_parameter_satisfy)
    #         found = False  # the function name
    #         pos_parameter = 0
    #         for i, line in enumerate(without_bracket_code):
    #             if function_name in line:
    #                 found = True
    #             if found and pos_parameter == functions.no_parameter_satisfy and not is_comment_line(line):
    #                 problem_lines.append(code[i])
    #             if not is_comment_line(line) and ',' in line:
    #                 pos_parameter += line.count(',')
    # else:
    #     if modeling_file:
    #         problem_lines.extend(modeling_file.code_context)
    #         noline = modeling_file.lineno
    #     else:
    #         problem_lines.extend(stack[2].code_context)
    #         noline = stack[2].lineno
    #     functions.no_parameter_satisfy = None
    #
    # s = "\n\tAt line " + str(stack[1].lineno) + " of function " + stack[1].function + " in file " + stack[1].filename[stack[1].filename.rindex(os.sep) + 1:]
    # s += "\n\tThis assertion fails: " + stack[1].code_context[0].strip()
    # if modeling_file is not None and len(problem_lines) > 0:
    #     if functions.no_parameter_satisfy is not None:
    #         s += "\n\n\tThe problem comes from parameter " + str(functions.no_parameter_satisfy) + " of satisfy() in file " \
    #              + modeling_file.filename[modeling_file.filename.rindex(os.sep) + 1:] + ","
    #     else:
    #         s += "\n\n\tThe problem is at line " + str(noline) + " of file " + modeling_file.filename[modeling_file.filename.rindex(os.sep) + 1:] + ","
    #     s += "\n\tWhen executing:\n\t\t" + "\n\t\t".join(problem_lines) + "\n\t" + message if message is not None else ""
    #
    #     # TODO get the good docstring
    #     # s += "\n\n\tDocumentation for function: " + stack[1].function
    #     # doc = docstringOf(stack[1].function)
    #     # s += doc if doc is not None else "\n\tNo python docstring for this function"

    raise TypeError(message)  # s)



"""
See https://en.wikipedia.org/wiki/Stable_roommates_problem
See "Stable Roommates and Constraint Programming" by Patrick Prosser. CPAIOR 2014: 15-28

Example of Execution:
  python3 RoomMate.py -data=RoomMate_sr0006.json
"""

from pycsp3 import *

preferences = data
n = len(preferences)


def pref_rank():
    pref = [[0] * n for _ in range(n)]  # pref[i][k] = j <-> ith guy has jth guy as kth choice
    rank = [[0] * n for _ in range(n)]  # rank[i][j] = k <-> ith guy ranks jth guy as kth choice
    for i in range(n):
        for k in range(len(preferences[i])):
            j = preferences[i][k] - 1  # because we start at 0
            rank[i][j] = k
            pref[i][k] = j
        rank[i][i] = len(preferences[i])
        pref[i][len(preferences[i])] = i
    return pref, rank


pref, rank = pref_rank()

# x[i] is the value of k, meaning that j = pref[i][k] is the paired agent
x = VarArray(size=n, dom=lambda i: range(len(preferences[i])))

satisfy(
    (imply(x[i] > rank[i][k], x[k] < rank[k][i]), imply(x[i] == rank[i][k], x[k] == rank[k][i])) for i in range(n) for k in pref[i] if k != i
)





"""
Problem 015 on CSPLib

The variant 'mod' corresponds to the one proposed in [Bessiere Meseguer Freuder Larrosa, On forward checking for non-binary constraint satisfaction, 2002].

Examples of Execution:
  python3 SchurrLemma.py
  python3 SchurrLemma.py -data=[10,10]
  python3 SchurrLemma.py -data=[10,10] -variant=mod
"""

from pycsp3 import *

n, d = data or (8, 8)  # n is the number of balls -- d is the number of boxes

# x[i] is the box where the ith ball is put
x = VarArray(size=n, dom=range(d))

if not variant():
    satisfy(
        NValues(x[i], x[j], x[k]) > 1 for (i, j, k) in product(range(n), repeat=3) if i < j and i + 1 + j == k
    )
elif variant("mod"):
    satisfy(
        AllDifferent(x[i], x[j], x[k]) for (i, j, k) in product(range(n), repeat=3) if i < j and i + 1 + j == k
    )



"""
Problem 024 on CSPLib

Examples of Execution:
  python3 Langford.py
  python3 Langford.py -data=[3,10]
"""

from pycsp3 import *

k, n = data or (3, 12)  # k is the number of occurrences for a value -- n is the number of values

# x[i][j] is the position in the sequence of the ith occurrence of j+1
x = VarArray(size=[k, n], dom=range(k * n))

satisfy(
    # all positions of the sequence must be set
    AllDifferent(x),

    # there are j numbers between two occurrences of j
    [x[i + 1][j] == x[i][j] + j + 2 for i in range(k - 1) for j in range(n)]
)



"""
See Challenge ROADEF 2001 (FAPP: Problème d'affectation de fréquences avec polarization)

Examples of Execution:
  python3 Fapp.py -data=Fapp_ex2.json
  python3 Fapp.py -data=Fapp_ex2.json -variant=short
"""

from pycsp3 import *

domains, routes, hard_constraints, soft_constraints = data
domains = [domains[route.domain] for route in routes]  # we skip the indirection
polarizations = [route.polarization for route in routes]
n, nSofts = len(routes), len(data.softs)


def table_soft(i, j, eq_relaxation, ne_relaxation, short_table=True):
    def calculate_size():
        for l in range(kl - 1):
            if distance >= t[l]:
                return l
        return kl - 1

    table = []  # we use a list instead of a set because is is quite faster to process
    cache = {}
    for f1 in domains[i]:
        for f2 in domains[j]:
            distance = abs(f1 - f2)
            key = str(distance) + " " + str(polarizations[i]) + " " + str(polarizations[j])
            if key not in cache:
                suffixes = []
                for pol in range(4):
                    p1 = 0 if pol < 2 else 1
                    p2 = 1 if pol in {1, 3} else 0
                    if (polarizations[i], p1) in [(1, 0), (-1, 1)] or (polarizations[j], p2) in [(1, 0), (-1, 1)]:
                        continue
                    t = eq_relaxation if p1 == p2 else ne_relaxation  # eqRelaxations or neRelaxations
                    for kl in range(12):
                        if kl == 11 or distance >= t[kl]:  # for kl=11, we suppose t[kl] = 0
                            suffixes.append((p1, p2, kl, 0 if kl == 0 or distance >= t[kl - 1] else 1, 0 if kl <= 1 else calculate_size()))
                cache[key] = suffixes
            elif short_table:
                continue
            for suffix in cache[key]:
                table.append((distance, *suffix) if short_table else (f1, f2, *suffix))
    return table


# f[i] is the frequency of the ith radio-link
f = VarArray(size=n, dom=lambda i: domains[i])

# p[i] is the polarization of the ith radio-link
p = VarArray(size=n, dom=lambda i: {0, 1} if polarizations[i] == 0 else {1} if polarizations[i] == 1 else {0})

# k is the relaxation level to be optimized
k = Var(dom=range(12))

# v1[q] is 1 iff the qth pair of radio-electric compatibility constraints is violated when relaxing another level
v1 = VarArray(size=nSofts, dom={0, 1})

# v2[q] is the number of times the qth pair of radio-electric compatibility constraints is violated when relaxing more than one level
v2 = VarArray(size=nSofts, dom=range(11))

satisfy(
    # imperative constraints
    dst == gap if eq else dst != gap for (dst, eq, gap) in [(abs(f[i] - f[j] if fq else p[i] - p[j]), eq, gap) for (i, j, fq, eq, gap) in hard_constraints]
)

if not variant():
    satisfy(
        # soft radio-electric compatibility constraints
        (f[i], f[j], p[i], p[j], k, v1[l], v2[l]) in table_soft(i, j, tuple(eqr), tuple(ner), False) for l, (i, j, eqr, ner) in enumerate(soft_constraints)
    )

elif variant("short"):
    soft_links = [[False] * n for _ in range(n)]
    for c in data.softs:
        soft_links[c.route1][c.route2] = soft_links[c.route2][c.route1] = True

    # d[i][j] is the distance between the ith and the jth frequencies (for i < j when a soft link exists)
    d = VarArray(size=[n, n], dom=lambda i, j: {abs(f1 - f2) for f1 in domains[i] for f2 in domains[j]} if i < j and soft_links[i][j] else None)

    satisfy(
        # computing intermediary distances
        [d[i][j] == abs(f[i] - f[j]) for i, j in combinations(range(n), 2) if d[i][j]],

        # soft radio-electric compatibility constraints
        [(d[min(i, j)][max(i, j)], p[i], p[j], k, v1[l], v2[l]) in table_soft(i, j, tuple(er), tuple(nr)) for l, (i, j, er, nr) in enumerate(soft_constraints)]
    )

minimize(
    k * (10 * nSofts ** 2) + Sum(v1) * (10 * nSofts) + Sum(v2)
)


# Note that
# a) we transform lists in tuples of relaxation arrays for speeding up calculations
# b) when gap is 0, abs(x - y) == gap (resp., abs(x - y) != gap) is automatically simplified into x == y (resp., x != y)



"""
One morning, in 1880, four families (the Boyds, Garveys, Logans and Navarros) stopped by Purdey's general store,.
Each family bought a different item: 50 pounds of flour, two gallons of kerosene, ten yards of muslin cloth, ten pounds of sugar.
One family paid cash, one took the item on credit and the other two traded other items for it (one a cured ham and the other a bushel of peas).
We know that:
 - the Boyds were new in town, and this was their first visit to the store
 - the family (which wasn't the Logans) that traded the bushel of peas didn't buy the kerosene
 - the Boyds and the Garveys bought the kerosene and the muslin in some order
 - one family traded a cured ham for a large sack of flour
 - Purdey only extended credit to regular customers, such as the family that bought the muslin on credit

Execution:
  python3 Purdey.py
"""

from pycsp3 import *

families = Boyds, Garveys, Logans, Navarros = "Boyds", "Garveys", "Logans", "Navarros"

flour = Var(families)
kerosene = Var(families)
cloth = Var(families)
sugar = Var(families)

cash = Var(families)
credit = Var(families)
ham = Var(families)
peas = Var(families)

satisfy(
    AllDifferent(flour, kerosene, cloth, sugar),
    AllDifferent(cash, credit, ham, peas),
    peas != Logans,
    peas != kerosene,
    (kerosene == Boyds) | (cloth == Boyds),
    (kerosene == Garveys) | (cloth == Garveys),
    ham == flour,
    credit == cloth,
    credit != Boyds
)



from pycsp3.problems.tests.tester import Tester, run

NAME = "g5_special"

diff = (Tester(NAME)
        )

same = (Tester(NAME)
        .add("DistinctVectors", data="[10,6,4]")
        .add("Domino", data="[200,200]")
        .add("Domino", data="[200,200]", variant="table")
        .add("Hanoi", data="5")
        .add("Knights", data="[16,4]")
        .add("Pigeons", data="6")
        .add("Pigeons", data="6", variant="dec")
        .add("Pions", data="[6,4]")
        .add("PropStress", data="20")
        .add("QueensPions", data="[6,4,3]")
        .add("QueensPions", data="[6,4,3]", variant="mul")
        )

xcsp = (Tester(NAME)
        .add("DistinctVectors", data="[10,6,4]")
        .add("Domino", data="[200,200]")
        .add("Domino", data="[200,200]", variant="table", options_py="-recognizeSlides")
        .add("Hanoi", data="5", options_py="-recognizeSlides")
        .add("Knights", data="[16,4]", options_py="-recognizeSlides")
        .add("Pigeons", data="6")
        .add("Pigeons", data="6", variant="dec")
        .add("Pions", data="[6,4]")
        .add("PropStress", data="20")
        .add("QueensPions", data="[6,4,3]")
        .add("QueensPions", data="[6,4,3]", variant="mul")
        )

run(xcsp,diff, same)



"""
Problem 032 on CSPLib

Examples of Execution:
  python3 StillLife.py
  python3 StillLife.py -data=[7,7]
  python3 StillLife.py -data=[7,7] -variant=wastage
"""

from pycsp3 import *

n, m = data or (8, 8)


if not variant():
    table = {(v, 0) for v in range(9) if v != 3} | {(2, 1), (3, 1)}


    def scope(i, j):
        return [x[k][l] for k in range(n) for l in range(m) if i - 1 <= k <= i + 1 and j - 1 <= l <= j + 1 and (k, l) != (i, j)]


    # x[i][j] is 1 iff the cell at row i and col j is alive
    x = VarArray(size=[n, m], dom={0, 1})

    # a[i][j] is the number of alive neighbours
    a = VarArray(size=[n, m], dom=range(9))

    satisfy(
        # computing the numbers of alive neighbours
        [Sum(scope(i, j)) == a[i][j] for i in range(n) for j in range(m)],

        # imposing rules of the game
        [(a[i][j], x[i][j]) in table for i in range(n) for j in range(m)],

        # imposing rules for ensuring valid dead cells around the board
        [
            [x[0][i:i + 3] != (1, 1, 1) for i in range(m - 2)],
            [x[- 1][i: i + 3] != (1, 1, 1) for i in range(m - 2)],
            [x[i:i + 3, 0] != (1, 1, 1) for i in range(n - 2)],
            [x[i:i + 3, - 1] != (1, 1, 1) for i in range(n - 2)]
        ],

        # tag(symmetry-breaking)
        (x[0][0] >= x[n - 1][n - 1], x[0][n - 1] >= x[n - 1][0]) if n == m else None
    )

    maximize(
        # maximizing the number of alive cells
        Sum(x)
    )

elif variant("wastage"):
    assert n == m


    def condition_for_tuple(t0, t1, t2, t3, t4, t5, t6, t7, t8, wa):
        s3 = t1 + t3 + t5 + t7
        s1 = t0 + t2 + t6 + t8 + s3
        s2 = t0 * t2 + t2 * t8 + t8 * t6 + t6 * t0 + s3
        return (t4 != 1 or (2 <= s1 <= 3 and (s2 > 0 or wa >= 2) and (s2 > 1 or wa >= 1))) and \
               (t4 != 0 or (s1 != 3 and (0 < s3 < 4 or wa >= 2)) and (s3 > 1 or wa >= 1))


    table = {(*t, i) for t in product(range(2), repeat=9) for i in range(3) if condition_for_tuple(*t, i)}

    #  x[i][j] is 1 iff the cell at row i and col j is alive (note that there is a border)
    x = VarArray(size=[n + 2, n + 2], dom=lambda i, j: {0} if i in {0, n + 1} or j in {0, n + 1} else {0, 1})

    #  w[i][j] is the wastage for the cell at row i and col j
    w = VarArray(size=[n + 2, n + 2], dom={0, 1, 2})

    # ws[i] is the wastage sum for cells at row i
    ws = VarArray(size=n + 2, dom=range(2 * (n + 2) * (n + 2) + 1))

    satisfy(
        # ensuring that cells at the border remain dead
        [
            [x[1][j:j + 3] != (1, 1, 1) for j in range(n)],
            [x[n][j:j + 3] != (1, 1, 1) for j in range(n)],
            [x[i:i + 3, 1] != (1, 1, 1) for i in range(n)],
            [x[i:i + 3, n] != (1, 1, 1) for i in range(n)]
        ],

        # still life + wastage constraints
        [(x[i - 1:i + 2, j - 1:j + 2], w[i][j]) in table for i in range(1, n + 1) for j in range(1, n + 1)],

        # managing wastage on the border
        [
            [(w[0][j] + x[1][j] == 1, w[n + 1][j] + x[n][j] == 1) for j in range(1, n + 1)],
            [(w[i][0] + x[i][1] == 1, w[i][n + 1] + x[i][n] == 1) for i in range(1, n + 1)],
        ],

        # summing wastage
        [Sum(w[0] if i == 0 else [ws[i - 1], w[i]]) == ws[i] for i in range(n + 2)],

        # tag(redundant-constraints)
        [ws[n + 1] - ws[i] >= 2 * ((n - i) // 3) + n // 3 for i in range(n + 1)]
    )

    maximize(
        # maximizing the number of alive cells
        (2 * n * n + 4 * n - ws[-1]) // 4
    )

# Note that:

# a) we could have posted unary constraints instead of identifying specific when creating arrays of variables:
# # cells at the border are assumed to be dead
# [
#     [(x[0][j] == 0, x[- 1][j] == 0) for j in range(n + 2)],
#     [(x[i][0] == 0, x[i][- 1] == 0) for i in range(n + 2)],
# ],

# b) in order to generate automatically slide constraints when handling rules for the boarder, we could
# post groups of constraints separately, i.e., write:
#
# # imposing rules for ensuring valid dead cells around the board
# [x[0][i:i + 3] != (1, 1, 1) for i in range(m - 2)],
#  ... (while using the option -recognizeSlides)

# c) we could use extension constraints instead of intension constraints, which would give:
# [x[0][i:i + 3] not in {(1, 1, 1)} for i in range(m - 2)],
# By the way, note that expressing such intension constraints are possible here because x[0][i:i + 3] has type 'ListVar'
# If it was not the case, as for example in: (x[0][0], x[0][1], x[0][2]) != (1,1,1), we would have to call cp_array
# which would give : cp_array(x[0][0], x[0][1], x[0][2]) != (1,1,1)



from pycsp3 import *


# TODO : in progress

n, prizes = data.n, data.prizes
prizes = cp_array(row + [0] for row in prizes)  # we add 0 for unused nodes; note that cp_array is required so as to be able to use the constraint 'element'

# s[i] is the node that succeeds to the ith node in the tour ('n' if unused)
s = VarArray(size=n, dom=range(n + 1))

# p[i] is the position of the ith node in the tour ('n' if unused)
p = VarArray(size=n + 1 if not variant() else n, dom=range(n + 1))

# g[i] is the gain collected when going from the ith node to its successor in the tour
g = VarArray(size=n, dom=lambda i: set(prizes[i]))

satisfy(
    # node 0 is the first node of the tour
    p[0] == 0,

    # managing unused nodes
    [iff(p[i] != n, s[i] != n) for i in range(n)],

    # each node appears at most once during the tour
    [Count(s, value=i) <= 1 for i in range(n)],

    # computing gains
    [prizes[i][s[i]] == g[i] for i in range(n)]
)

if not variant():
    # z[i] is the position of the successor of the ith node in the tour
    z = VarArray(size=n, dom=range(n + 1))


    def tab():
        return {(0, ANY, 0), (n, n, n)} | {(i, j, j + 1) for i in range(1, n) for j in range(0, n - 1)}


    print(sorted(list(tab())))

    satisfy(
        p[n] == n,

        #        [(s[i], p[i], z[i]) in tab() for i in range(n)],

        [imply(p[i] == n - 1, z[i] != n) for i in range(n)],

        [z[i] == ift(s[i] == 0, 0, ift(s[i] == n, n, p[i] + 1)) for i in range(n)],

        [p[s[i]] == z[i] for i in range(n)]
    )


elif variant("table"):
    def table(pos):
        tbl = {(0,) + (ANY,) * n, (n,) + (ANY,) * n}  # tuple([-1] + [ANY] * n), tuple([0] + [ANY] * n)
        for vi in range(n):
            if vi != pos and (vi != 0 or pos == 0):
                for vv in range(1, n):
                    tbl.add((vi,) + tuple(vv if i == vi else vv - 1 if i == pos else ANY for i in range(n)))
        return tbl


    satisfy(
        s[0] == 1,
        s[1] == 2,
        s[2] == 3,
        s[3] == 4,

        # linking variables from s and p
        [(s[i], *p) in table(i) for i in range(n)]
    )

maximize(
    # maximizing the sum of collected gains
    Sum(g)
)



# [imply(s[i] > 0, p[s[i]] == p[i]+1) for i in range(n)],  # TODO alternative to the short table above. How to do that ? meta-constraint ifThen ? reification ?



import re
from enum import unique

from pycsp3.classes.auxiliary.ptypes import auto, AbstractType, TypeXML
from pycsp3.classes.main.constraints import ConstraintUnmergeable
from pycsp3.classes.main.variables import Variable
from pycsp3.tools.inspector import checkType
from pycsp3.tools.utilities import flatten


@unique
class TypeVarHeuristic(AbstractType):
    LEXICO, DOM, DEG, DDEG, WDEG, IMPACT, ACTIVITY = auto(7)


@unique
class TypeValHeuristic(AbstractType):
    CONFLICTS, VALUE = auto(2)


@unique
class TypeConsistency(AbstractType):
    FC, BC, AC, SAC, FPWC, PC, CDC, FDAC, EDAC, VAC = auto(10)

    def __str__(self):
        return self.name


@unique
class TypeBranching(AbstractType):
    TWO_WAY, D_WAY = auto(2)

    def __str__(self):
        return self.name.replace("_", "-").replace("TWO", "2").lower()


@unique
class TypeRestart(AbstractType):
    LUBY, GEOMETRIC = auto(2)


@unique
class TypeAnnArg(AbstractType):
    TYPE = auto()
    STATIC, RANDOM, MIN, MAX = auto(4)
    LC = auto()
    ORDER = auto()
    CONSISTENCY, BRANCHING, CUTOFF, FACTOR = auto(4)
    START_INDEX, START_ROW_INDEX, START_COL_INDEX = auto(3)


class Annotation(ConstraintUnmergeable):
    pass


class AnnotationDecision(Annotation):
    def __init__(self, variables):
        super().__init__(TypeXML.DECISION)
        variables = flatten(variables)
        checkType(variables, [Variable])
        self.arg(TypeXML.DECISION, variables)


class AnnotationOutput(Annotation):
    def __init__(self, variables):
        super().__init__(TypeXML.OUTPUT)
        variables = flatten(variables)
        checkType(variables, [Variable])
        self.arg(TypeXML.OUTPUT, variables)


class AnnotationHeuristic(Annotation):
    def __init__(self, name):
        super().__init__(name)

    # To keep the good order
    def add_arguments(self, random_part, min_part, max_part):
        if random_part:
            self.arg(TypeAnnArg.RANDOM, random_part[0] if random_part[0] else [None])
        if min_part:
            self.arg(TypeAnnArg.MIN, min_part[0] if min_part[0] else [None], attributes=[(TypeAnnArg.TYPE, min_part[1])])
        if max_part:
            self.arg(TypeAnnArg.MAX, max_part[0] if max_part[0] else [None], attributes=[(TypeAnnArg.TYPE, max_part[1])])


class AnnotationVarHeuristic(AnnotationHeuristic):
    def __init__(self, h):
        super().__init__(TypeXML.VAR_HEURISTIC)
        checkType(h, VarHeuristic)
        self.attributes.append((TypeAnnArg.LC, h.lc))
        if h.staticData:
            self.arg(TypeAnnArg.STATIC, h.staticData)
        self.add_arguments(h.randomPart, h.minPart, h.maxPart)


class AnnotationValHeuristic(AnnotationHeuristic):
    def __init__(self, h):
        super().__init__(TypeXML.VAL_HEURISTIC)
        checkType(h, ValHeuristic)
        if h.staticData:
            self.arg(TypeAnnArg.STATIC, h.staticData[0], attributes=[(TypeAnnArg.ORDER, " ".join(str(ele) for ele in h.staticData[1]))])
        self.add_arguments(h.randomData, h.minData, h.maxData)


class AnnotationFiltering(Annotation):
    def __init__(self, consistency):
        super().__init__(TypeXML.FILTERING)
        checkType(consistency, TypeConsistency)
        self.attributes.append((TypeAnnArg.TYPE, consistency))


class AnnotationPrepro(Annotation):
    def __init__(self, consistency):
        super().__init__(TypeXML.PREPRO)
        checkType(consistency, TypeConsistency)
        self.attributes.append((TypeAnnArg.CONSISTENCY, consistency))


class AnnotationSearch(Annotation):
    def __init__(self, search):
        super().__init__(TypeXML.SEARCH)
        checkType(search, Search)
        self.attributes = [(TypeAnnArg.CONSISTENCY, search.consistency), (TypeAnnArg.BRANCHING, search.branching)]


class AnnotationRestarts(Annotation):
    def __init__(self, restarts):
        super().__init__(TypeXML.RESTARTS)
        checkType(restarts, Restarts)
        self.attributes = [(TypeAnnArg.TYPE, restarts.type), (TypeAnnArg.CUTOFF, restarts.cutoff), (TypeAnnArg.FACTOR, restarts.factor)]


''' Annotations classes '''


class Search:
    def __init__(self, *, consistency=None, branching=None):
        assert consistency or branching
        assert isinstance(consistency, (TypeConsistency, type(None))) and isinstance(branching, (TypeBranching, type(None)))
        self.consistency = consistency
        self.branching = branching


class Restarts:
    def __init__(self, *, type, cutoff, factor=1):
        assert isinstance(type, TypeRestart) and isinstance(cutoff, int) and isinstance(factor, float)
        self.type = type
        self.cutoff = cutoff
        self.factor = factor


class VHeuristic:
    def __init__(self):
        self.staticPart = None
        self.randomPart = None
        self.minPart = None
        self.maxPart = None

    def random(self, variables=None):
        variables = flatten(variables)
        checkType(variables, ([Variable], type(None)))
        self.randomPart = (variables,)
        return self

    def _opt(self, variables, type):
        if variables:
            variables = flatten(variables)
            checkType(variables, [Variable])
        types = TypeVarHeuristic if isinstance(self, VarHeuristic) else TypeValHeuristic
        assert isinstance(type, str) and all(p in [t.name for t in types] for p in re.split(r'/|\+', type)), "Bad value for " + type
        return variables, type

    def min(self, variables=None, *, type):
        self.minPart = self._opt(variables, type)
        return self

    def max(self, variables=None, *, type):
        self.maxPart = self._opt(variables, type)
        return self


class VarHeuristic(VHeuristic):
    def __init__(self, *, lc=None):
        super().__init__()
        self.lc = lc

    def static(self, variables):
        variables = flatten(variables)
        checkType(variables, [Variable])
        self.staticPart = variables
        return self


class ValHeuristic(VHeuristic):
    def __init__(self):
        super().__init__()

    def static(self, variables, *, order):
        variables = flatten(variables)
        checkType(variables, [Variable])
        order = flatten(order)
        checkType(order, [int])
        self.staticPart = (variables, order)
        return self



"""
See Problem in MiniZinc -- https://github.com/MiniZinc/minizinc-benchmarks/tree/master/amaze

Given a grid containing pairs of numbers (ranging from 1 to a greater value), connect the pairs (e.g. 1 to 1; 2 to 2; etc)
by drawing a line horizontally and vertically, but not diagonally.
The lines must never cross.

Example of Execution:
  python3 Amaze.py -data=Amaze_simple.json
"""

from pycsp3 import *

n, m, points1, points2 = data
nPairs = len(points1)

# x[i][j] is the value at row i and column j (a boundary is put around the board).
x = VarArray(size=[n + 2, m + 2], dom=lambda i, j: {0} if i in {0, n + 1} or j in {0, m + 1} else range(nPairs + 1))

table = ({(0, ANY, ANY, ANY, ANY)}
         | {tuple(ne(v) if k in (i, j) else v for k in range(5)) for i, j in combinations(range(1, 5), 2) for v in range(1, nPairs + 1)})

satisfy(
    # putting two occurrences of each value on the board
    [x[i, j] == v + 1 for v in range(nPairs) for (i, j) in [points1[v], points2[v]]],

    # each cell with a fixed value has exactly one neighbour with the same value
    [Count([x[i - 1][j], x[i + 1][j], x[i][j - 1], x[i][j + 1]], value=v + 1) == 1 for v in range(nPairs) for (i, j) in [points1[v], points2[v]]],

    # each empty cell either contains 0 or has exactly two neighbours with the same value
    [(x[i][j], x[i - 1][j], x[i + 1][j], x[i][j - 1], x[i][j + 1]) in table for i in range(1, n + 1) for j in range(1, m + 1) if
     [i, j] not in points1 + points2]
)

minimize(
    Sum(x)
)


# Note that

# a) the table contains (smart) conditions, which makes code more compact than:
# table = ({(0, ANY, ANY, ANY, ANY)}
#          | {(v, v, v, ne(v), ne(v)) for v in range(1, nPairs + 1)}
#          | {(v, v, ne(v), v, ne(v)) for v in range(1, nPairs + 1)}
#          | {(v, v, ne(v), ne(v), v) for v in range(1, nPairs + 1)}
#          | {(v, ne(v), v, v, ne(v)) for v in range(1, nPairs + 1)}
#          | {(v, ne(v), v, ne(v), v) for v in range(1, nPairs + 1)}
#          | {(v, ne(v), ne(v), v, v) for v in range(1, nPairs + 1)})

# b) even if  data come from a text file via a parser that builds tuples (and not lists)
#    we have to write [i, j] not in points1 + points2 because tuples (in data) are automatically converted to lists



""""
See https://ipg.host.cs.st-andrews.ac.uk/challenge/

Examples of Execution:
  python3 OpenStacks.py -data=OpenStacks_example.json -variant=m1
  python3 OpenStacks.py -data=OpenStacks_example.json -variant=m2
"""

from pycsp3 import *

orders = data
n, m = len(orders), len(orders[0])  # n orders (customers), m possible products

if variant("m1"):
    def table1(i):
        v = sum(orders[i])
        return {(0, 0, 0)} | {(i, ANY, 1) for i in range(1, v)} | {(v, 0, 0), (v, 1, 1)}


    # p[j] is the period (time) of the jth product
    p = VarArray(size=m, dom=range(m))

    # np[i][j] is the number of products made at time j and required by customer i
    np = VarArray(size=[n, m], dom=lambda i, j: range(sum(orders[i]) + 1))

    # r[i][t] is 1 iff the product made at time t concerns customer i
    r = VarArray(size=[n, m], dom={0, 1})

    # o[i][t] is 1 iff the stack is open for customer i at time t
    o = VarArray(size=[n, m], dom={0, 1})

    satisfy(
        # all products are scheduled at different times
        AllDifferent(p),

        [orders[i][p[j]] == r[i][j] for i in range(n) for j in range(m)],

        [np[i][j] == (r[i][j] if j == 0 else np[i][j - 1] + r[i][j]) for i in range(n) for j in range(m)],

        [(np[i][j], r[i][j], o[i][j]) in table1(i) for i in range(n) for j in range(m)],
    )

    minimize(
        # minimizing the number of stacks that are simultaneously open
        Maximum(Sum(o[:, t]) for t in range(m))
    )

elif variant("m2"):
    def table2(t):
        return {(ANY, te, 0) for te in range(t)} | {(ts, ANY, 0) for ts in range(t + 1, m)} | {(ts, te, 1) for ts in range(t + 1) for te in range(t, m)}


    # p[j] is the period (time) of the jth product
    p = VarArray(size=m, dom=range(m))

    # s[i] is the starting time of the ith stack
    s = VarArray(size=n, dom=range(m))

    # e[i] is the ending time of the ith stack
    e = VarArray(size=n, dom=range(m))

    # o[i][t] is 1 iff the ith stack is open at time t
    o = VarArray(size=[n, m], dom={0, 1})

    satisfy(
        # all products are scheduled at different times
        AllDifferent(p),

        # computing starting times of stacks
        [Minimum(p[j] for j in range(m) if orders[i][j] == 1) == s[i] for i in range(n)],

        # computing ending times of stacks
        [Maximum(p[j] for j in range(m) if orders[i][j] == 1) == e[i] for i in range(n)],

        # inferring when stacks are open
        [(s[i], e[i], o[i][t]) in table2(t) for i in range(n) for t in range(m)],
    )

    minimize(
        # minimizing the number of stacks that are simultaneously open
        Maximum(Sum(o[:, t]) for t in range(m))
    )


# to have ordinary tables, we have to use: to_ordinary_table(tab, [v + 1, 2, 2]) and to_ordinary_table(tab, [m, m, 2])

# If we want explicitly the number of open stacks at time t, we write instead:
# ns[t] is the number of open stacks at time t
# ns = VarArray(size=m, dom=range(m + 1))
# computing the number of open stacks at any time
# [Sum(o[:, j]) == ns[j] for j in range(m)]
# minimize (Maximum(ns))



"""
Problem 022 on CSPLib

The model below is without the cost parameter, as in Minizinc.

Example of Execution:
  python3 BusScheduling.py -data=BusScheduling_r1.json
"""

from pycsp3 import *

nTasks, shifts = data
nShifts = len(shifts)

# x[i] is 1 iff the ith shift is selected
x = VarArray(size=nShifts, dom={0, 1})

satisfy(
    # each task is covered by exactly one shift
    Count(x[i] for i, shift in enumerate(shifts) if t in shift) == 1 for t in range(nTasks)
)

minimize(
    # minimizing the number of shifts
    Sum(x)
)


# Note that the default value for Count is 1. Wxe can equivalently write:
# Count([x[i] for i, shift in enumerate(shifts) if t in shift], value=1) == 1 for t in range(nTasks)



from pycsp3.problems.tests.tester import Tester, run

NAME = "g1_single"

diff = (Tester(NAME)
        .add("Allergy")
        .add("Alpha")
        .add("Picnic")
        .add("Purdey")
        .add("Recipe")
        .add("Sandwich")
        .add("Zebra")
        )

same = (Tester(NAME)
        .add("LabeledDice")
        .add("TrafficLights")
        .add("Witch")
        )

xcsp = (Tester(NAME)
        .add("Agatha")
        .add("Allergy")
        .add("Alpha")
        .add("Alpha", variant="var")
        .add("LabeledDice")
        .add("NFractions")
        .add("Photo")
        .add("Photo", variant="aux")
        .add("Picnic")
        .add("Purdey")
        .add("Recipe")
        .add("Sandwich")
        .add("SendMore")
        .add("TrafficLights")
        .add("Witch")
        .add("Zebra")
        )

run(xcsp, diff, same)



"""
See https://www.researchgate.net/publication/220270875_The_Traveling_Tournament_Problem_Description_and_Benchmarks

Examples of Execution:
  python3 TravelingTournament.py -data=TravelingTournament_galaxy04.json -variant=a2
  python3 TravelingTournament.py -data=TravelingTournament_galaxy04.json -variant=a3
"""

from pycsp3 import *

distances = data
nTeams, nRounds = len(distances), len(distances) * 2 - 2
assert nTeams % 2 == 0, "An even number of teams is expected"
nConsecutiveGames = 2 if variant("a2") else 3  # used in one comment


def table_end(i):
    # note that when playing at home (whatever the opponent, travel distance is 0)
    return {(1, ANY, 0)} | {(0, j, distances[i][j]) for j in range(nTeams) if j != i}


def table_intern(i):
    return ({(1, 1, ANY, ANY, 0)} |
            {(0, 1, j, ANY, distances[i][j]) for j in range(nTeams) if j != i} |
            {(1, 0, ANY, j, distances[i][j]) for j in range(nTeams) if j != i} |
            {(0, 0, j1, j2, distances[j1][j2]) for j1 in range(nTeams) for j2 in range(nTeams) if different_values(i, j1, j2)})


def automaton():
    tr2 = [("q", 0, "q01"), ("q", 1, "q11"), ("q01", 0, "q02"), ("q01", 1, "q11"), ("q11", 0, "q01"), ("q11", 1, "q12"), ("q02", 1, "q11"), ("q12", 0, "q01")]
    tr3 = [("q02", 0, "q03"), ("q12", 1, "q13"), ("q03", 1, "q11"), ("q13", 0, "q01")]
    fi2 = {"q01", "q02", "q11", "q12"}
    fi3 = {"q03", "q13"}
    return Automaton(start="q", final=fi2 if variant("a2") else fi2 | fi3, transitions=tr2 if variant("a2") else tr2 + tr3)


# o[i][k] is the opponent (team) of the ith team  at the kth round
o = VarArray(size=[nTeams, nRounds], dom=range(nTeams))

# h[i][k] is 1 iff the ith team plays at home at the kth round
h = VarArray(size=[nTeams, nRounds], dom={0, 1})

# a[i][k] is 0 iff the ith team plays away at the kth round
a = VarArray(size=[nTeams, nRounds], dom={0, 1})

# t[i][k] is the travelled distance by the ith team at the kth round. An additional round is considered for returning at home.
t = VarArray(size=[nTeams, nRounds + 1], dom={d for row in distances for d in row})

satisfy(

    # each team must play exactly two times against each other team
    [Cardinality(o[i], occurrences={j: 2 for j in range(nTeams) if j != i}, closed=True) for i in range(nTeams)],

    # ensuring symmetry of games: if team i plays against j at round k, then team j plays against i at round k
    [o[o[i][k]][k] == i for i in range(nTeams) for k in range(nRounds)],

    # playing home at round k iff not playing away at round k
    [h[i][k] == ~a[i][k] for i in range(nTeams) for k in range(nRounds)],

    # channeling the three arrays
    [h[o[i][k]][k] == a[i][k] for i in range(nTeams) for k in range(nRounds)],

    # playing against the same team must be done once at home and once away
    [imply(o[i][k1] == o[i][k2], h[i][k1] != h[i][k2]) for i in range(nTeams) for k1, k2 in combinations(range(nRounds), 2)],

    # at each round, opponents are all different  tag(redundant-constraints)
    [AllDifferent(o[:, k]) for k in range(nRounds)],

    # tag(symmetry-breaking)
    o[0][0] < o[0][- 1],

    # at most 'nConsecutiveGames' consecutive games at home, or consecutive games away
    [h[i] in automaton() for i in range(nTeams)],

    # handling travelling for the first game
    [(h[i][0], o[i][0], t[i][0]) in table_end(i) for i in range(nTeams)],

    # handling travelling for the last game
    [(h[i][- 1], o[i][- 1], t[i][-1]) in table_end(i) for i in range(nTeams)],

    # handling travelling for two successive games
    [(h[i][k], h[i][k + 1], o[i][k], o[i][k + 1], t[i][k + 1]) in table_intern(i) for i in range(nTeams) for k in range(nRounds - 1)]

)

minimize(
    # minimizing summed up travelled distance
    Sum(t)
)



"""
See https://en.wikipedia.org/wiki/Kakuro
See "Kakuro as a Constraint Problem" by Helmut Simonis

Example of Execution:
  python3 Kakuro.py -data=Kakuro_easy-000.json
"""

from pycsp3 import *

nRows, nCols, clues = data

# x[i][j] is the value put at row i and column j
x = VarArray(size=[nRows, nCols], dom=lambda i, j: range(1, 10) if clues[i][j].x == clues[i][j].y == 0 else None)


def structures():
    def h_scope(i, j):
        assert clues[i][j].x > 0
        t = []
        for k in range(j + 1, nCols):
            if clues[i][k].x != 0:
                break
            t.append(x[i][k])
        return t

    def v_scope(i, j):
        assert clues[i][j].y > 0
        t = []
        for k in range(i + 1, nRows):
            if clues[k][j].y != 0:
                break
            t.append(x[k][j])
        return t

    h = [(h_scope(i, j), clues[i][j].x) for i in range(nRows) for j in range(nCols) if clues[i][j].x > 0]
    v = [(v_scope(i, j), clues[i][j].y) for i in range(nRows) for j in range(nCols) if clues[i][j].y > 0]
    return h, v


horizontal, vertical = structures()

if not variant():
    satisfy(
        [Sum(scp) == clue for (scp, clue) in horizontal],

        [AllDifferent(scp) for (scp, _) in horizontal],

        [Sum(scp) == clue for (scp, clue) in vertical],

        [AllDifferent(scp) for (scp, _) in vertical]
    )

elif variant("table"):
    cache = dict()


    def table(limit, arity):  # tuples with different values summing to the specified limit
        n_values, offset = 9, 1  # hard coding for this context
        key = str(limit) + "_" + str(n_values) + "_" + str(arity) + "_" + str(offset)
        if key in cache:
            return cache[key]
        tuples = set()
        combs = list(combinations(range(n_values), arity))
        perms = list(permutations(range(arity)))
        for comb in combs:
            if offset != 0:
                comb = [v + offset for v in comb]
            if sum(comb) == limit:
                for perm in perms:
                    tuples.add(tuple(comb[perm[i]] for i in range(arity)))
        cache[key] = tuples
        return tuples


    satisfy(
        [scp in table(clue, len(scp)) for (scp, clue) in horizontal],

        [scp in table(clue, len(scp)) for (scp, clue) in vertical]
    )



from pycsp3 import *

#  comment discarded
x, y, z = VarArray(size=3, dom={0, 1})

#  comment not discarded
xx = Var(dom={0, 1})

yy = Var(dom=[0, 1])
try:
    zz = Var(dom=(0, 1))
except AssertionError as e:
    print("Test assert domain with tuple 1:", e)

try:
    # comment discarded
    g1 = VarArray(size=[e for e in range(10)], dom=range(100))
except AssertionError as e:
    print("Test assert dimensions != 0:", e)

# comment not discarded
g1 = VarArray(size=[e for e in range(1, 5)], dom=range(100))
g2 = VarArray(size=[e for e in range(1, 5)], dom=[1, 2, 3])
g3 = VarArray(size=[e for e in range(1, 5)], dom={1, 2, 3})

try:
    g3 = VarArray(size=[e for e in range(1, 5)], dom=(1, 2, 3))
except AssertionError as e:
    print("Test assert domain with tuple 2:", e)

z1 = VarArray(size=[2] * 10, dom=range(100))

try:
    z1 = VarArray(size="tata", dom=range(100))
except TypeError as e:
    print("Type error in size:", e)

try:
    z1 = VarArray(size=1, dom="g")
except TypeError as e:
    print("Type error in domain:", e)

try:
    z1 = VarArray(dom="g")
except TypeError as e:
    print("Missing a parameter:", e)

try:
    z1 = VarArray(size=1)
except TypeError as e:
    print("Missing a parameter:", e)

try:
    z1 = Var()
except TypeError as e:
    print("Missing a parameter:", e)

try:
    VarArray(size=[2], dom={0, 1})
except AssertionError as e:
    print("Not assigned VarArray:", e)

try:
    Var(dom={0, 1})
except AssertionError as e:
    print("Not assigned Var:", e)



"""
All squares of a board of a specified size (specified numbers of rows and columns) must be colored with the minimum number of colors.
The four corners of any rectangle inside the board must not be assigned the same color.

Examples of Execution:
  python3 BoardColoration.py
  python3 BoardColoration.py -data=[8,10]
"""

from pycsp3 import *

n, m = data or (6, 5)

# x[i][j] is the color at row i and column j
x = VarArray(size=[n, m], dom=range(n * m))

satisfy(
    # at least two corners of different colors for any rectangle inside the board
    [NValues(x[i1][j1], x[i1][j2], x[i2][j1], x[i2][j2]) > 1 for i1, i2 in combinations(range(n), 2) for j1, j2 in combinations(range(m), 2)],

    # tag(symmetry-breaking)
    LexIncreasing(x, matrix=True)
)

minimize(
    # minimizing the greatest used color index (and, consequently, the number of colors)
    Maximum(x)
)



"""
Problem 017 on CSPLib

The edges of a complete graph (with n nodes) must be coloured with the minimum number of colours.
There must be no monochromatic triangle in the graph, i.e. in any triangle at most two edges have the same colour.
With 3 colours, the problem has a solution if n < 17.

Examples of Execution:
  python3 Ramsey.py
  python3 Ramsey.py -data=10
"""

from pycsp3 import *

n = data or 8

# x[i][j] is the color of the edge between nodes i and j
x = VarArray(size=[n, n], dom=lambda i, j: range((n * (n - 1)) // 2) if i < j else None)

satisfy(
    # no monochromatic triangle in the graph
    NValues(x[i][j], x[i][k], x[j][k]) > 1 for (i, j, k) in combinations(range(n), 3)
)

minimize(
    Maximum(x)
)



from pycsp3.problems.data.parsing import *

data["mapSize"] = next_int()
nRadars = next_int()
data["maxCoverage"] = next_int()
nInsignificantCells = next_int()
series = next_int()
seed = next_int()


# TODO to be finished (as in JvSCP3)


"""
See https://turing.cs.hbg.psu.edu/txn131/graphcoloring.html

Examples of Execution:
  python3 GraphColoring.py -data=GraphColoring_1-fullins-3.json
  python3 GraphColoring.py -data=GraphColoring_1-fullins-3.json -variant=sum
"""

from pycsp3 import *

n, edges, colorings, multiColorings = data  # n is the number of nodes -- multi-coloring not taken into account for the moment
colorings = colorings if colorings else []

# c[i] is the color assigned to the ith node
c = VarArray(size=n, dom=range(n))

satisfy(
    [abs(c[i] - c[j]) >= d for (i, j, d) in edges],

    # nodes with preassigned colors
    [c[i] == colors[0] for (i, colors) in colorings if len(colors) == 1],

    # nodes with subsets of prefixed colors
    [c[i] in colors for (i, colors) in colorings if len(colors) > 1]
)

if not variant():
    minimize(
        # minimizing the greatest used color index (and, consequently, the number of colors)
        Maximum(c)
    )
elif variant("sum"):
    minimize(
        # minimizing the sum of colors assigned to nodes
        Sum(c)
    )


# Note that
# a) when d is 1, abs(x - y) >= d is automatically simplified into x != y



from pycsp3.problems.data.parsing import *

n = next_int()
nClues = next_int()

clues = [[0 for _ in range(n)] for _ in range(n)]
for _ in range(nClues):
    x, y, z = next_int() - 1, next_int() - 1, next_int()
    clues[x][y] = z

data["n"] = n
data["clues"] = clues



"""
Problem 007 on CSPLib

Examples of Execution:
  python3 AllInterval.py
  python3 AllInterval.py -data=10
  python3 AllInterval.py -data=10 -variant=aux
"""

from pycsp3 import *

n = data or 8

# x[i] is the ith note of the series
x = VarArray(size=n, dom=range(n))

if not variant():
    satisfy(
        # notes must occur once, and so form a permutation
        AllDifferent(x),

        # intervals between neighbouring notes must form a permutation
        AllDifferent(abs(x[i + 1] - x[i]) for i in range(n - 1)),

        # tag(symmetry-breaking)
        x[0] < x[n - 1]
    )

elif variant("aux"):
    # y[i] is the distance between x[i] and x[i+1]
    y = VarArray(size=n - 1, dom=range(1, n))

    satisfy(
        # notes must occur once, and so form a permutation
        AllDifferent(x),

        # intervals between neighbouring notes must form a permutation
        AllDifferent(y),

        # computing distances
        [y[i] == abs(x[i + 1] - x[i]) for i in range(n - 1)],

        # tag(symmetry-breaking)
        [x[0] < x[n - 1], y[0] < y[1]]
    )



"""
Problem 026 on CSPLib

Examples of Execution:
  python3 SportsScheduling.py
  python3 SportsScheduling.py -data=10
  python3 SportsScheduling.py -data=10 -variant=dummy
"""

from pycsp3 import *

nTeams = data or 8
nWeeks, nPeriods, nMatches = nTeams - 1, nTeams // 2, (nTeams - 1) * nTeams // 2


def match_number(t1, t2):
    return nMatches - ((nTeams - t1) * (nTeams - t1 - 1)) // 2 + (t2 - t1 - 1)


table = {(t1, t2, match_number(t1, t2)) for t1, t2 in combinations(range(nTeams),2)}

# h[w][p] is the home team at week w and period p
h = VarArray(size=[nWeeks, nPeriods], dom=range(nTeams))

# a[w][p] is the away team at week w and period p
a = VarArray(size=[nWeeks, nPeriods], dom=range(nTeams))

# m[w][p] is the number of the match at week w and period p
m = VarArray(size=[nWeeks, nPeriods], dom=range(nMatches))

satisfy(
    # linking variables through ternary table constraints
    [(h[w][p], a[w][p], m[w][p]) in table for w in range(nWeeks) for p in range(nPeriods)],

    # all matches are different (no team can play twice against another team)
    AllDifferent(m),

    # each week, all teams are different (each team plays each week)
    [AllDifferent(h[w] + a[w]) for w in range(nWeeks)],

    # each team plays at most two times in each period
    [Cardinality(h[:, p] + a[:, p], occurrences={t: range(1, 3) for t in range(nTeams)}) for p in range(nPeriods)],

    # tag(symmetry-breaking)
    [
        # the match '0 versus t' (with t strictly greater than 0) appears at week t-1
        [Count(m[w], value=match_number(0, w + 1)) == 1 for w in range(nWeeks)],

        # the first week is set : 0 vs 1, 2 vs 3, 4 vs 5, etc.
        [m[0][p] == match_number(2 * p, 2 * p + 1) for p in range(nPeriods)]
    ]
)

if variant("dummy"):
    # hd[p] is the home team for the dummy match of period p  tag(dummy-week)
    hd = VarArray(size=nPeriods, dom=range(nTeams))

    # ad[p] is the away team for the dummy match of period p  tag(dummy-week)
    ad = VarArray(size=nPeriods, dom=range(nTeams))

    satisfy(
        # handling dummy week (variables and constraints)  tag(dummy-week)
        [
            # all teams are different in the dummy week
            AllDifferent(hd + ad),

            # each team plays two times in each period
            [Cardinality(h[:, p] + a[:, p] + [hd[p], ad[p]], occurrences={t: 2 for t in range(nTeams)}) for p in range(nPeriods)],

            # tag(symmetry-breaking)
            [hd[p] < ad[p] for p in range(nPeriods)]
        ]
    )



"""
See https://en.wikipedia.org/wiki/Sudoku
See, e.g., "Sudoku as a Constraint Problem" by Helmut Simonis

Example of Execution:
  python3 Sudoku.py -data=[9,None]
  python3 Sudoku.py -data=Sudoku_s13a.json
  python3 Sudoku.py -data=Sudoku_s13a.json -variant=table
  python3 Sudoku.py -data=Sudoku_example.txt -dataparser=Sudoku_Parser.py
"""

import math

from pycsp3 import *

n, clues = data  # n (order of the grid) is typically 9 -- if not 0, clues[i][j] is a value imposed at row i and col j
base = int(math.sqrt(n))
assert base * base == n

# x[i][j] is the value in cell at row i and col j.
x = VarArray(size=[n, n], dom=range(1, n + 1))

if not variant():
    satisfy(
        # imposing distinct values on each row and each column
        AllDifferent(x, matrix=True),

        # imposing distinct values on each block  tag(blocks)
        [AllDifferent(x[i:i + base, j:j + base]) for i in range(0, n, base) for j in range(0, n, base)]
    )

elif variant("table"):
    table = list(permutations(range(1, n + 1)))
    satisfy(
        # imposing distinct values on each row and each column
        [line in table for line in x + columns(x)],

        # imposing distinct values on each block  tag(blocks)
        [x[i:i + base, j:j + base] in table for i in range(0, n, base) for j in range(0, n, base)]
    )

satisfy(
    # imposing clues  tag(clues)
    x[i][j] == clues[i][j] for i in range(n) for j in range(n) if clues and clues[i][j] > 0
)


# Note that :

# a) using set(permutations(range(1, n + 1))) instead of list(permutations(range(1, n + 1))) is far less time efficient



""""
Problem 001 on CSPLib

Examples of Execution:
  python3 CarSequencing.py -data=CarSequencing_dingbas.json
  python3 CarSequencing.py -data=CarSequencing_dingbas.json -variant=table
"""

from pycsp3 import *

classes, limits = data
nCars, nClasses, nOptions = sum(cl.demand for cl in classes), len(classes), len(limits)


def redu(k, i):
    # i stands for the number of blocks set to the maximal capacity
    remaining = sum(options[k] * demand for (demand, options) in classes) - i * limits[k].num
    possible = nCars - i * limits[k].den
    if remaining > 0 and possible > 0:
        return Sum(o[j][k] for j in range(possible)) >= remaining


# c[i] is the class of the ith assembled car
c = VarArray(size=nCars, dom=range(nClasses))

# o[i][k] is 1 if the ith assembled car has option k
o = VarArray(size=[nCars, nOptions], dom={0, 1})

satisfy(
    # building the right numbers of cars per class
    Cardinality(c, occurrences={i: classes[i].demand for i in range(nClasses)})
)

if not variant():
    satisfy(
        # constraints about options
        imply(c[i] == j, o[i][k] == options[k]) for i in range(nCars) for j, (_, options) in enumerate(classes) for k in range(nOptions)
    )
elif variant("table"):
    satisfy(
        # constraints about options
        (c[i], *o[i]) in {(j, *options) for j, (_, options) in enumerate(classes)} for i in range(nCars)
    )

satisfy(
    # constraints about option frequencies
    [Sum(o[i:i + den, k]) <= num for k, (num, den) in enumerate(limits) for i in range(nCars) if i <= nCars - den],

    # tag(redundant-constraints)
    [redu(k, i) for k in range(nOptions) for i in range(nCars)]
)



from pycsp3.tools.curser import queue_in


class Diagram:
    def __init__(self, transitions):
        self.transitions = Diagram._add_transitions(transitions)

    def __contains__(self, other):
        queue_in.append((self, other))
        return True

    MSG_STATE = "states must given under the form of strings"

    @staticmethod
    def _add_transitions(transitions):
        assert len(transitions) > 0, "at least one transition must be present"
        assert isinstance(transitions, (list, set))
        t = []
        for transition in transitions:
            if isinstance(transition, list):
                transition = tuple(transition)
            assert isinstance(transition, tuple), "A transition must be given under the form of a 3-tuple (or a list)"
            assert len(transition) == 3, "Error: each transition must be composed of 3 elements"
            state1, state2 = transition[0], transition[2]
            assert isinstance(state1, str) and isinstance(state2, str), Diagram.MSG_STATE
            values = transition[1] if isinstance(transition[1], (list, tuple, set, range)) else [transition[1]]
            for value in values:
                assert isinstance(value, int), "currently, the value of a transition is necessarily an integer"
                t.append((state1, value, state2))
        return t


class Automaton(Diagram):
    @staticmethod
    def q(i, j=None):
        return "q" + str(i) + ("" if j is None else "x" + str(j))

    def __init__(self, *, start, transitions, final):
        super().__init__(transitions)
        self.start = start
        self.final = [final] if isinstance(final, str) else sorted(set(final))
        assert isinstance(self.start, str) and all(isinstance(f, str) for f in self.final), Diagram.MSG_STATE


class MDD(Diagram):
    def __init__(self, transitions):
        assert isinstance(transitions, list)  # currently, a list is wanted for a MDD (and not a set); to be changed?
        super().__init__(transitions)



from pycsp3.problems.data.parsing import *
from pycsp3.compiler import Compilation
import random


def random_edge():
    x, y = random.randint(0, data.nNodes - 1), random.randint(0, data.nNodes - 1)
    return (x, y) if x != y else random_edge()


data.nNodes = ask_number("Number of nodes ?")
data.nColors = ask_number("Number of colors ?")
nEdges = ask_number("Number of edges ?")
data.edges = [random_edge() for _ in range(nEdges)]

Compilation.string_data = "-" + "-".join(str(v) for v in (data.nNodes, data.nColors, nEdges))



from pycsp3 import *

# Problem 028 at CSPLib

r = data.r

# x[i][j] is the value of the matrix at row i and column j
x = VarArray(size=[r, r], dom={0, 1})

satisfy(
    # constraints on rows
    [Sum(row) in range(r // 2, r//2+1) for row in x],

    # constraints on columns
    [Sum(col) in range(r // 2, r//2 +1) for col in columns(x)],

    LexIncreasing(x, matrix=True)

)



"""
See https://en.wikipedia.org/wiki/Latin_square

Example of Execution:
  python3 LatinSquare2.py -data=LatinSquare2_7-2-0.json
"""

from pycsp3 import *

n, clues = data  # clues are given by tuples of the form (row, col, value)

# x[i][j] is the value at row i and col j of the Latin Square
x = VarArray(size=[n, n], dom=range(n))

satisfy(
    AllDifferent(x, matrix=True),

    # tag(clues)
    [x[i][j] == v for (i, j, v) in clues] if clues else None,

    # tag(diagonals)
    [AllDifferent(dgn) for dgn in diagonals_down(x, broken=True) + diagonals_up(x, broken=True)]
)



from pycsp3.problems.tests.tester import Tester, run

NAME = "g4_world"

diff = (Tester(NAME)
        .add("Bacp", data="Bacp_10.json", variant="m1")
        .add("Bacp", data="Bacp_10.json", variant="m2")
        .add("Bacp", data="Bacp_10.json", variant="m1-d")
        .add("Bacp", data="Bacp_10.json", variant="m2-d")
        .add("Bacp", data="Bacp_10.mzn", variant="m1", prs_py="Bacp_ParserZ", prs_jv="Bacp_ParserZ")  # m1 is enough to test parsing
        .add("CarSequencing", data="CarSequencing_dingbas.json")
        .add("CarSequencing", data="CarSequencing_dingbas.json", variant="table")
        .add("League", data="League_010-03-04.json")
        .add("NurseRostering", data="NurseRostering_00.json")
        .add("NurseRostering", data="NurseRostering_18.json")
        .add("PizzaVoucher", data="PizzaVoucher_pizza6.json")
        .add("Rack", data="Rack_r2.json")
        .add("Rack2", data="Rack_r2b.json")
        .add("RadarSurveillance", data="RadarSurveillance_8-24-3-2-00.json")
        .add("SolitaireBattleship", data="SolitaireBattleship_sb-12-12-5-0.json")  # 54 solutions
        .add("SolitaireBattleship", data="SolitaireBattleship-battleship_instances-00113.json")  # different order of groups in a block
        .add("SteelMillSlab", data="SteelMillSlab_bench_2_0.json")
        .add("SteelMillSlab", data="SteelMillSlab_bench_2_0.json", variant="01")
        .add("TemplateDesign", data="TemplateDesign_catfood_2.json")
        .add("TemplateDesign", data="TemplateDesign_catfood_2.json", variant="aux")
        .add("Tal", data="Tal-frobserved-7-15-11-13-9-1-11-7-4_1.json")
        .add("TravelingPurchaser", data="TravelingPurchaser-7-5-30-1.json")  # intension constraints more compact in PyCSP
        .add("Vrp", data="Vrp_P-n16-k8.json")
        .add("Warehouse", data="Warehouse_example.txt", prs_py="Warehouse_Parser", prs_jv="Warehouse_Parser")
        .add("Wwtpp", data="Wwtpp_ex04400.json")
        .add("Wwtpp", data="Wwtpp_ex04400.json", variant="short")
        )

same = (Tester(NAME)
        .add("Crossword", data="Crossword_vg6-7-ogd.json")
        .add("Crossword", data="Crossword_h1501-lex.json")
        .add("Crossword", data="Crossword_h0504-lex.json", variant="alt")
        .add("Fapp", data="Fapp_ex2.json")
        .add("Fapp", data="Fapp_ex2.json", variant="short")
        # .add("Fapp", data="Fapp_01-0200.json")  # TODO too long (and seems to crash)
        # .add("Fapp", data="Fapp_01-0200.json", variant="short")
        .add("QuadraticAssignment", data="QuadraticAssignment_qap.json")
        .add("QuadraticAssignment", data="QuadraticAssignment_example.txt", prs_py="QuadraticAssignment_Parser", prs_jv="QuadraticAssignment_Parser")
        .add("MisteryShopper", data="MisteryShopper_04.json")
        .add("Rcpsp", data="Rcpsp_j30-01-01.json")
        .add("Rlfap", data="Rlfap_card-scen-04.json", variant="card")
        .add("Rlfap", data="Rlfap_span-scen-05.json", variant="span")
        .add("SchedulingFS", data="Taillard-fs-020-05-1.json")
        .add("SchedulingFS", data="SchedulingFS-Taillard-os-04-04-0.json")
        .add("SchedulingJS", data="Taillard-js-015-15-1.json")
        .add("SchedulingJS", data="Sadeh-js-e0ddr1-0.json")
        .add("SchedulingOS", data="Taillard-os-05-05-1.json")
        .add("SchedulingOS", data="GP-os-01.json")
        .add("TravelingTournament", data="TravelingTournament_galaxy04.json", variant="a2")
        .add("TravelingTournament", data="TravelingTournament_galaxy04.json", variant="a3")
        .add("TravelingTournamentWithPredefinedVenues", data="Ttppv_circ8bbal.json", variant="a2")
        .add("TravelingTournamentWithPredefinedVenues", data="Ttppv_circ8bbal.json", variant="a3")
        )

xcsp = (Tester(NAME)
        .add("Bacp", data="Bacp_10.json", variant="m1")  # optimum 26
        .add("Bacp", data="Bacp_10.json", variant="m2")  # optimum 26
        .add("Bacp", data="Bacp_10.json", variant="m1-d")  # optimum 1
        .add("Bacp", data="Bacp_10.json", variant="m2-d")  # optimum 1
        .add("Bacp", data="Bacp_10.mzn", variant="m1", prs_py="Bacp_ParserZ", prs_jv="Bacp_ParserZ")  # m1 is enough to test parsing; optimum 26
        .add("CarSequencing", data="CarSequencing_dingbas.json")  # 6 solutions
        .add("CarSequencing", data="CarSequencing_dingbas.json", variant="table")  # 6 solutions
        .add("Crossword", data="Crossword_vg6-7-ogd.json")  # sat
        .add("Crossword", data="Crossword_h1501-lex.json")  # sat
        .add("Crossword", data="Crossword_h0504-lex.json", variant="alt")
        .add("Fapp", data="Fapp_ex2.json")  # optimum 13871
        .add("Fapp", data="Fapp_ex2.json", variant="short")  # optimum 13871
        # .add("Fapp", data="Fapp_01-0200.json")  # TODO too long
        # .add("Fapp", data="Fapp_01-0200.json", variant="short")  # around 22 seconds for generating the file
        .add("League", data="League_010-03-04.json")  # optimum 92
        .add("MisteryShopper", data="MisteryShopper_04.json")  # 501552 solutions
        .add("NurseRostering", data="NurseRostering_00.json")  # optimum 1202
        .add("NurseRostering", data="NurseRostering_18.json")
        .add("PizzaVoucher", data="PizzaVoucher_pizza6.json")  # optimum 210
        .add("QuadraticAssignment", data="QuadraticAssignment_qap.json")  # optimum 4776
        .add("QuadraticAssignment", data="QuadraticAssignment_example.txt", prs_py="QuadraticAssignment_Parser",
             prs_jv="QuadraticAssignment_Parser")  # optimum 4776
        .add("Rack", data="Rack_r2.json")  # optimum 1100
        .add("Rack2", data="Rack_r2b.json")  # optimum 1100
        .add("RadarSurveillance", data="RadarSurveillance_8-24-3-2-00.json")  # sat
        .add("Rcpsp", data="Rcpsp_j30-01-01.json")
        .add("Rlfap", data="Rlfap_card-scen-04.json", variant="card")  # optimum 46
        .add("Rlfap", data="Rlfap_span-scen-05.json", variant="span")  # optimum 792
        .add("Rlfap", data="Rlfap_max-graph-05.json", variant="max")  # optimum 221  (use -ub=222 with AbsCon to prove it, or -os=increasing)
        .add("SchedulingFS", data="SchedulingFS-Taillard-os-04-04-0.json")  # optimum 302
        .add("SchedulingFS", data="Taillard-fs-020-05-1.json")
        .add("SchedulingJS", data="Sadeh-js-e0ddr1-0.json")
        .add("SchedulingJS", data="Taillard-js-015-15-1.json")
        .add("SchedulingOS", data="Taillard-os-05-05-1.json")
        .add("SchedulingOS", data="GP-os-01.json")
        .add("SolitaireBattleship", data="SolitaireBattleship_sb-12-12-5-0.json")  # 51 solutions
        .add("SolitaireBattleship", data="SolitaireBattleship-battleship_instances-00113.json")  # 1 solution
        .add("SteelMillSlab", data="SteelMillSlab_bench_2_0.json")
        .add("SteelMillSlab", data="SteelMillSlab_bench_2_0.json", variant="01")
        .add("Tal", data="Tal-frobserved-7-15-11-13-9-1-11-7-4_1.json")  # optimum 142
        .add("TemplateDesign", data="TemplateDesign_catfood_2.json")  # optimum 2 ; java AbsCon TemplateDesign-catfood_2.xml -valh=Rand -p=SAC3 -sop
        .add("TemplateDesign", data="TemplateDesign_catfood_2.json", variant="aux")
        .add("TravelingPurchaser", data="TravelingPurchaser-7-5-30-1.json")  # optimum 124
        .add("TravelingTournament", data="TravelingTournament_galaxy04.json", variant="a2")  # optimum 517
        .add("TravelingTournament", data="TravelingTournament_galaxy04.json", variant="a3")  # optimum 416
        .add("TravelingTournamentWithPredefinedVenues", data="Ttppv_circ8bbal.json", variant="a2")  # optimum 94
        .add("TravelingTournamentWithPredefinedVenues", data="Ttppv_circ8bbal.json", variant="a3")  # optimum 80
        .add("Vrp", data="Vrp_P-n16-k8.json")
        .add("Warehouse", data="Warehouse_example.txt", prs_py="Warehouse_Parser", prs_jv="Warehouse_Parser")  # optimum 383
        .add("Wwtpp", data="Wwtpp_ex04400.json")  # unsat
        .add("Wwtpp", data="Wwtpp_ex04400.json", variant="short")  # unsat
        )

run(xcsp, diff, same)



from pycsp3.problems.data.parsing import *

next_int()  # nItems
next_int()  # nBids
data["bids"] = [OrderedDict([("value", t[0]), ("items", [int(t[i]) for i in range(1, len(t))])]) for line in remaining_lines() for t in [line.split(" ")]]



"""
See https://en.wikipedia.org/wiki/Job_shop_scheduling

Example of Execution:
  python3 SchedulingJS.py -data=Sadeh-js-e0ddr1-0.json
"""
from pycsp3 import *


jobs = data
horizon = max(job.dueDate for job in jobs) if all(job.dueDate != -1 for job in jobs) else sum(sum(job.durations) for job in jobs)
durations = [job.durations for job in jobs]
indexes = [[job.resources.index(j) for j in range(len(job.durations))] for job in jobs]
n, m = len(jobs), len(jobs[0].durations)

# s[i][j] is the start time of the jth operation for the ith job
s = VarArray(size=[n, m], dom=range(horizon))

satisfy(
    # operations must be ordered on each job
    [Increasing(s[i], lengths=durations[i]) for i in range(n)],

    # respecting release dates
    [s[i][0] > jobs[i].releaseDate for i in range(n) if jobs[i].releaseDate > 0],

    # respecting due dates
    [s[i][- 1] <= jobs[i].dueDate - durations[i][- 1] for i in range(n) if 0 <= jobs[i].dueDate < horizon - 1],

    # no overlap on resources
    [NoOverlap(origins=[s[i][indexes[i][j]] for i in range(n)], lengths=[durations[i][indexes[i][j]] for i in range(n)]) for j in range(m)]
)

minimize(
    # minimizing the makespan
    Maximum(s[i][- 1] + durations[i][- 1] for i in range(n))
)



"""
Problem 063 on CSPLib

Examples of Execution:
  python3 Auction.py
  python3 Auction.py -data=Auction_example.json
"""

from pycsp3 import *

bids = data or default_data("Auction_example.json")
items = sorted({item for bid in bids for item in bid.items})
values = integer_scaling(bid.value for bid in bids)
nBids = len(bids)

# x[i] is 1 iff the ith bid is selected
x = VarArray(size=nBids, dom={0, 1})

satisfy(
    # avoiding intersection of bids
    Count(scp, value=1) <= 1 for scp in [[x[i] for i, bid in enumerate(bids) if item in bid.items] for item in items]
)

maximize(
    # maximizing summed values of selected bids
    x * values
)



"""
Problem 030 on CSPLib

Examples of Execution:
  python3 Bacp.py -data=Bacp_10.json -variant=m1
  python3 Bacp.py -data=Bacp_10.json -variant=m2
  python3 Bacp.py -data=Bacp_10.json -variant=m1-d
  python3 Bacp.py -data=Bacp_10.json -variant=m2-d
"""

from pycsp3 import *

nCourses, nPeriods, minCredits, maxCredits, minCourses, maxCourses, credits, prerequisites = data
maxCredits = maxCredits * maxCourses if subvariant("d") else maxCredits
assert nCourses == len(credits)

# s[c] is the period (schedule) for course c
s = VarArray(size=nCourses, dom=range(nPeriods))

# co[p] is the number of courses at period p
co = VarArray(size=nPeriods, dom=range(minCourses, maxCourses + 1))

# cr[p] is the number of credits at period p
cr = VarArray(size=nPeriods, dom=range(minCredits, maxCredits + 1))

if variant("m1"):
    def table(c):
        return {(0,) * p + (credits[c],) + (0,) * (nPeriods - p - 1) + (p,) for p in range(nPeriods)}


    # cp[c][p] is 0 if the course c is not planned at period p, the number of credits for c otherwise
    cp = VarArray(size=[nCourses, nPeriods], dom=lambda c, p: {0, credits[c]})

    satisfy(
        # channeling between arrays cp and s
        [(*cp[c], s[c]) in table(c) for c in range(nCourses)],

        # counting the number of courses in each period
        [Count(s, value=p) == co[p] for p in range(nPeriods)],

        # counting the number of credits in each period
        [Sum(cp[:, p]) == cr[p] for p in range(nPeriods)]
    )

elif variant("m2"):
    # pc[p][c] is 1 iff the course c is at period p
    pc = VarArray(size=[nPeriods, nCourses], dom={0, 1})

    satisfy(
        # tag(channeling)
        [iff(pc[p][c], s[c] == p) for p in range(nPeriods) for c in range(nCourses)],

        # ensuring that each course is assigned to a period
        [Sum(pc[:, c]) == 1 for c in range(nCourses)],

        # counting the number of courses in each period
        [Sum(pc[p]) == co[p] for p in range(nPeriods)],

        # counting the number of credits in each period
        [pc[p] * credits == cr[p] for p in range(nPeriods)]
    )

satisfy(
    # handling prerequisites
    s[c1] < s[c2] for (c1, c2) in prerequisites
)

if subvariant("d"):
    minimize(
        # minimizing the maximal distance in term of credits
        Maximum(cr) - Minimum(cr)
    )
else:
    minimize(
        # minimizing the maximum number of credits in periods
        Maximum(cr)
    )

annotate(decision=s)


# Note that:

# a) the way we build the table could also be written as:
# return {tuple(credits[c] if j == p else p if j == nPeriods else 0 for j in range(nPeriods + 1)) for p in range(nPeriods)}



from pycsp3 import *

x = VarArray(size=[10, 10], dom=range(150))
y = VarArray(size=[10, 10], dom=range(150))
z = VarArray(size=[10], dom=range(150))

satisfy(
    #  Equal to instantiation
    [x[i][j] == i * 10 + j for i in range(10) for j in range(10)]
)

satisfy(
    min(abs(x[0][0]), abs(y[0][0])),
    max(abs(x[0][0]), min(y[0], z[1]))
)

try:
    # TODO
    satisfy(x == z)
except TypeError as e:
    print("TODO:", e)

try:
    # TODO
    satisfy(x[0] == z)
except AssertionError as e:
    print("TODO:", e)

satisfy(x[0][0] == y[0][0])

try:
    # TODO
    satisfy(x[0] in {0, 1})
except AssertionError as e:
    print("TODO bousille forbidden fruit:", e)

print("HERE")

satisfy(
    #  right
    [x[0][0] == 0,
     x[0][0] > 0,
     x[0][0] < 0,
     x[0][0] != 0,
     x[0][0] >= 0,
     x[0][0] <= 0,
     x[0][0] in {0, 1},
     x[0][0] not in {0, 1}]
)



"""
Problem 013 on CSPLib

Example of Execution:
  python3 ProgressiveParty.py -data=ProgressiveParty_example.json
"""

from pycsp3 import *

nPeriods, boats = data
nBoats = len(boats)
capacities, crews = zip(*boats)

# h[b] indicates if the boat b is a host boat
h = VarArray(size=nBoats, dom={0, 1})

# s[b][p] is the scheduled (visited) boat by the crew of boat b at period p
s = VarArray(size=[nBoats, nPeriods], dom=range(nBoats))

# g[b1][p][b2] is 1 if s[b1][p] = b2
g = VarArray(size=[nBoats, nPeriods, nBoats], dom={0, 1})

satisfy(
    # identifying host boats (when receiving)
    [iff(s[b][p] == b, h[b]) for b in range(nBoats) for p in range(nPeriods)],

    # identifying host boats (when visiting)
    [imply(s[b1][p] == b2, h[b2]) for b1 in range(nBoats) for b2 in range(nBoats) if b1 != b2 for p in range(nPeriods)],

    # channeling variables from arrays s and g
    [Channel(g[b][p], s[b][p]) for b in range(nBoats) for p in range(nPeriods)],

    # boat capacities must be respected
    [g[:, p, b] * crews <= capacities[b] for b in range(nBoats) for p in range(nPeriods)],

    # a guest boat cannot revisit a host
    [AllDifferent(s[b], excepting=b) for b in range(nBoats)],

    # guest crews cannot meet more than once
    [Sum(s[b1][p] == s[b2][p] for p in range(nPeriods)) <= 2 for b1, b2 in combinations(range(nBoats), 2)]
)

minimize(
    # minimizing the number of host boats
    Sum(h)
)


# less compact way of posting : [[g[i][p][b] for i in range(nBoats)] * crews <= boats[b].capacity for b in range(nBoats) for p in range(nPeriods)],



from pycsp3 import *
from pycsp3.solvers.chocosolver import ChocoProcess, ChocoPy4J

n = 8

# x[i] is the ith note of the series
x = VarArray(size=n, dom=range(n))

satisfy(
    # notes must occur once, and so form a permutation
    AllDifferent(x),

    # intervals between neighbouring notes must form a permutation
    AllDifferent(abs(x[i + 1] - x[i]) for i in range(n - 1)),

    # tag(symmetry-breaking)
    x[0] < x[n - 1]
)


print("Compile:\n")
instance = compile()

print("\nStatic solving:\n")
solution = ChocoProcess().solve(instance)
print("solution:", solution)

print("\nPy4j solving:\n")
solver = ChocoPy4J()
solver.loadXCSP3(instance)

print("in progress")



"""
See Problem 038 on CSPLib

Examples of Execution:
  python3 SteelMillSlab.py -data=SteelMillSlab_bench_2_0.json
  python3 SteelMillSlab.py -data=SteelMillSlab_bench_2_0.json -variant=01
"""

from pycsp3 import *

capacities, orders = data
possibleLosses = [min(v for v in [0] + capacities if v >= i) - i for i in range(max(capacities) + 1)]
sizes, colors = zip(*orders)
allColors = sorted(set(colors))
colorGroups = [[i for i, order in enumerate(orders) if order.color == color] for color in allColors]
nOrders, nSlabs, nColors = len(orders), len(orders), len(allColors)

# sb[i] is the slab used to produce the ith order
sb = VarArray(size=nOrders, dom=range(nSlabs))

# ld[j] is the load of the jth slab
ld = VarArray(size=nSlabs, dom=range(max(capacities) + 1))

# ls[j] is the loss of the jth slab
ls = VarArray(size=nSlabs, dom=set(possibleLosses))

if not variant():
    satisfy(
        # computing (and checking) the load of each slab
        [[sb[i] == j for i in range(nOrders)] * sizes == ld[j] for j in range(nSlabs)],

        # computing the loss of each slab 
        [(ld[j], ls[j]) in {(i, loss) for i, loss in enumerate(possibleLosses)} for j in range(nSlabs)],

        # no more than two colors for each slab 
        [Sum(disjunction(sb[i] == j for i in g) for g in colorGroups) <= 2 for j in range(nSlabs)]
    )

elif variant("01"):
    # y[j][i] is 1 iff the jth slab is used to produce the ith order
    y = VarArray(size=[nSlabs, nOrders], dom={0, 1})

    # z[j][c] is 1 iff the jth slab is used to produce an order of color c
    z = VarArray(size=[nSlabs, nColors], dom={0, 1})

    satisfy(
        # linking variables sb and y
        [iff(sb[i] == j, y[j][i]) for j in range(nSlabs) for i in range(nOrders)],

        # linking variables sb and z
        [imply(sb[i] == j, z[j][allColors.index(orders[i].color)]) for j in range(nSlabs) for i in range(nOrders)],

        # computing (and checking) the load of each slab
        [y[j] * sizes == ld[j] for j in range(nSlabs)],

        # computing the loss of each slab
        [(ld[j], ls[j]) in {(i, loss) for i, loss in enumerate(possibleLosses)} for j in range(nSlabs)],

        # no more than two colors for each slab
        [Sum(z[j]) <= 2 for j in range(nSlabs)]
    )

satisfy(
    # tag(redundant-constraints)
    Sum(ld) == sum(sizes),

    # tag(symmetry-breaking)
    [
        Decreasing(ld),

        [sb[i] <= sb[j] for i, j in combinations(range(nOrders), 2) if orders[i] == orders[j]]
    ]
)

minimize(
    # minimizing summed up loss
    Sum(ls)
)


# Note that:

# a) for computing (and checking) the load of each slab
#    the reverse side ? eq(z[s]{c] = 1 => or(...) isnot really necessary but could it be helpful?
